"""Implement a Quantum Convolutional Neural Network (QCNN) for classification tasks.
The QCNN model was originally proposed in: arXiv:1810.03787 <https://arxiv.org/abs/1810.03787>_
for the identification of quantum phases.

The QuantumCNN class in this module provides methods to construct the QCNN.
"""

import numpy as np

from qibo import gates
from qibo.backends import get_backend
from qibo.models import Circuit


class QuantumCNN:
    """
    Model that implements and trains a variational quantum convolutional network (QCNN) for
    classification tasks.
    The QCNN model was originally proposed in: `arXiv:1810.03787
    <https://arxiv.org/abs/1810.03787>`_ for the identification of quantum phases.

    Args:
        nqubits (int): number of qubits of the input states. Currently supports powers of 2.
        nlayers (int): number of convolutional and pooling layers.
        nclasses (int): number of classes to be classified. Default setting of 2 (phases).
        params: initial list of variational parameters. If not provided, all parameters
            will be initialized to zero.
        twoqubitansatz (:class:`qibo.models.Circuit`): a two qubit ansatz that can be input by
            the user to form the two qubit ansatz used in the convolutional circuit.

    Example:
        .. testcode::

            import math
            import random

            import numpy as np

            from qibo import set_backend
            from qibo.models.qcnn import QuantumCNN

            set_backend("numpy")

            data = np.random.rand(16)
            data = data / np.linalg.norm(data)
            data = [data]
            labels = [[1]]
            testbias = np.zeros(1)
            testangles = [random.uniform(0, 2 * np.pi) for i in range(21 * 2)]
            init_theta = np.concatenate((testbias, testangles))
            test_qcnn = QuantumCNN(nqubits=4, nlayers=1, nclasses=2, params=init_theta)
            testcircuit = test_qcnn._circuit
            result = test_qcnn.minimize(
               init_theta, data=data, labels=labels, nshots=10000, method="Powell"
            )
            predict0 = test_qcnn.predict(data, nshots=10000)

    """

    def __init__(
        self,
        nqubits,
        nlayers,
        nclasses=2,
        params=None,
        twoqubitansatz=None,
        copy_init_state=None,
    ):
        """
        Initializes the QuantumCNN object.

        Args:
            nqubits (int): The number of qubits in the QCNN.
            nlayers (int): The number of layers in the QCNN.
            nclasses (int, optional): The number of classes for the classification task. Defaults to 2.
            params (np.ndarray, optional): The initial parameters for the QCNN. If None, random parameters are generated. Defaults to None.
            twoqubitansatz (qibo.models.circuit.Circuit, optional): A two-qubit ansatz for the convolutional layers. If None, a default ansatz is used. Defaults to None.
            copy_init_state (bool, optional): Whether to copy the initial state for each shot in the simulation. If None, the behavior depends on the backend. Defaults to None.

        Raises:
            ValueError: If nqubits is not larger than 1.
        """

        self.nclasses = nclasses
        self.nqubits = nqubits
        self.nlayers = nlayers
        self.twoqubitansatz = twoqubitansatz

        if copy_init_state is None:
            if "qibojit" in str(get_backend()):
                self.copy_init_state = True
            else:
                self.copy_init_state = False

        if self.twoqubitansatz is None:
            self.nparams_conv = 15
        else:
            self.nparams_conv = len(self.twoqubitansatz.get_parameters())

        self.nparams_pool = 6
        self.nparams_layer = self.nparams_conv + self.nparams_pool
        self.measured_qubits = int(np.ceil(np.log2(self.nclasses)))

        if self.nqubits <= 1:
            raise ValueError("nqubits must be larger than 1")

        self._circuit = self.ansatz(nlayers, params=params)

    def quantum_conv_circuit(self, bits, symbols):
        """
        Internal helper function to construct a single convolutional layer.

        Args:
            bits: list or numpy.array with the qubits that the convolutional layer should apply to.
            symbols: list or numpy.array with the angles to be used in the circuit.
        Returns:
            Circuit for a single convolutional layer
        """
        circuit = Circuit(self.nqubits)
        for first, second in zip(bits[0::2], bits[1::2]):
            circuit += self.two_qubit_unitary([first, second], symbols)

        # check that there are more than 2 qubits to prevent double conv
        if len(bits) > 2:
            for first, second in zip(bits[1::2], bits[2::2] + [bits[0]]):
                circuit += self.two_qubit_unitary([first, second], symbols)
        return circuit

    def quantum_pool_circuit(self, source_bits, sink_bits, symbols):
        """
        Internal helper function to construct a single pooling layer.

        Args:
            source_bits: list or numpy.array with the source qubits for the pooling layer.
            sink_bits: list or numpy.array with the sink qubits for the pooling layer.
            symbols: list or numpy.array with the angles to be used in the circuit.
        Returns:
            Circuit for a single pooling layer
        """
        circuit = Circuit(self.nqubits)
        for source, sink in zip(source_bits, sink_bits):
            circuit += self.two_qubit_pool(source, sink, symbols)
        return circuit

    def ansatz(self, nlayers, params=None):
        """
        Args:
            theta: list or numpy.array with the angles to be used in the circuit.
            nlayers: int number of layers of the varitional circuit ansatz.
        Returns:
            Circuit implementing the QCNN variational ansatz.
        """

        nparams_conv = self.nparams_conv
        nparams_layer = self.nparams_layer

        if params is not None:
            symbols = params
        else:
            symbols = [0 for _ in range(nlayers * nparams_layer)]

        nbits = self.nqubits

        qubits = [_ for _ in range(nbits)]
        circuit = Circuit(self.nqubits)
        for layer in range(nlayers):
            conv_start = int(nbits - nbits / (2**layer))
            pool_start = int(nbits - nbits / (2 ** (layer + 1)))
            param_start = layer * nparams_layer
            circuit += self.quantum_conv_circuit(
                qubits[conv_start:], symbols[param_start : param_start + nparams_conv]
            )
            circuit += self.quantum_pool_circuit(
                qubits[conv_start:pool_start],
                qubits[pool_start:],
                symbols[param_start + nparams_conv : param_start + nparams_layer],
            )

        # Measurements
        circuit.add(gates.M(*[nbits - 1 - i for i in range(self.measured_qubits)]))

        return circuit

    def one_qubit_unitary(self, bit, symbols):
        """
        Internal helper function to make a circuit enacting a rotation of the bloch sphere about the X,
        Y and Z axis, that depends on the values in `symbols`.

        Args:
            bit: the qubit to apply the one-qubit unitaries to
            symbols: length 3 array containing the parameters
        Returns:
            Circuit containing the unitaries added to the specified qubit.
        """
        circuit = Circuit(self.nqubits)
        circuit.add(gates.RX(bit, symbols[0]))
        circuit.add(gates.RY(bit, symbols[1]))
        circuit.add(gates.RZ(bit, symbols[2]))

        return circuit

    def two_qubit_unitary(self, bits, symbols):
        """
        Internal helper function to create a circuit consisting of two qubit unitaries.

        Args:
            bits: the two qubits to apply the unitaries to
            symbols: length 15 array containing the parameters
        Returns:
            Circuit containing the unitaries added to the specified qubits.
        """

        circuit = Circuit(self.nqubits)

        if self.twoqubitansatz is None:
            circuit += self.one_qubit_unitary(bits[0], symbols[0:3])
            circuit += self.one_qubit_unitary(bits[1], symbols[3:6])
            circuit.add(gates.RZZ(bits[0], bits[1], symbols[6]))
            circuit.add(gates.RYY(bits[0], bits[1], symbols[7]))
            circuit.add(gates.RXX(bits[0], bits[1], symbols[8]))

            circuit += self.one_qubit_unitary(bits[0], symbols[9:12])
            circuit += self.one_qubit_unitary(bits[1], symbols[12:])

        else:
            circuit.add(self.twoqubitansatz.on_qubits(bits[0], bits[1]))
            circuit.set_parameters(symbols[0 : self.nparams_conv])

        return circuit

    def two_qubit_pool(self, source_qubit, sink_qubit, symbols):
        """
        Internal helper function to create a circuit to do a parameterized 'pooling' operation with controlled unitaries, which
        attempts to reduce entanglement down from two qubits to just one.

        Args:
            source_qubit: the control qubit.
            sink_qubit: the target qubit for the controlled unitaries.
            symbols: array with 6 elements containing the parameters.
        Returns:
            Circuit containing the unitaries added to the specified qubits.
        """
        pool_circuit = Circuit(self.nqubits)
        sink_basis_selector = self.one_qubit_unitary(sink_qubit, symbols[0:3])
        source_basis_selector = self.one_qubit_unitary(source_qubit, symbols[3:6])
        pool_circuit += sink_basis_selector
        pool_circuit += source_basis_selector
        pool_circuit.add(gates.CNOT(source_qubit, sink_qubit))
        pool_circuit += sink_basis_selector.invert()

        return pool_circuit

    def set_circuit_params(self, angles, has_bias=False):
        """
        Sets the parameters of the QCNN circuit. Can be used to load previously saved or optimized parameters.

        Args:
            angles: the parameters to be loaded.
            has_bias: specify whether the list of angles contains the bias.

        """
        if not has_bias:
            params = list(angles)
        else:
            self._optimal_angles = angles
            params = list(angles[self.measured_qubits :])

        expanded_params = []
        nbits = self.nqubits
        for layer in range(self.nlayers):
            nleft = nbits / (2**layer)
            param_start = layer * self.nparams_layer
            conv_params = params[param_start : param_start + self.nparams_conv]
            pool_params = params[
                param_start + self.nparams_conv : param_start + self.nparams_layer
            ]
            pool_params += [-pool_params[2], -pool_params[1], -pool_params[0]]
            expanded_params += conv_params * int(nleft if nleft > 2 else 1)
            expanded_params += pool_params * int(nleft / 2)

        self._circuit.set_parameters(expanded_params)

    def Classifier_circuit(self, theta):
        """
        Args:
            theta: list or numpy.array with the biases and the angles to be used in the circuit.
        Returns:
            Circuit implementing the variational ansatz for angles "theta".
        """
        bias = np.array(theta[0 : self.measured_qubits])
        angles = theta[self.measured_qubits :]

        self.set_circuit_params(angles)
        return self._circuit

    def Predictions(self, circuit, theta, init_state, nshots=10000):
        """
        Args:
            theta: list or numpy.array with the biases to be used in the circuit.
            init_state: numpy.array with the quantum state to be classified.
            nshots: int number of runs of the circuit during the sampling process (default=10000).
        Returns:
            numpy.array() with predictions for each qubit, for the initial state.
        """
        bias = np.array(theta[0 : self.measured_qubits])
        if self.copy_init_state:
            init_state_copy = init_state.copy()
        else:
            init_state_copy = init_state
        circuit_exec = circuit(init_state_copy, nshots)
        result = circuit_exec.frequencies(binary=False)
        prediction = np.zeros(self.measured_qubits)

        for qubit in range(self.measured_qubits):
            for clase in range(self.nclasses):
                binary = bin(clase)[2:].zfill(self.measured_qubits)
                prediction[qubit] += result[clase] * (1 - 2 * int(binary[-qubit - 1]))

        return prediction / nshots + bias

    def square_loss(self, labels, predictions):
        """
        Args:
            labels: list or numpy.array with the qubit labels of the quantum states to be classified.
            predictions: list or numpy.array with the qubit predictions for the quantum states to be classified.
        Returns:
            numpy.float32 with the value of the square-loss function.
        """
        loss = 0
        for l, p in zip(labels, predictions):
            for qubit in range(self.measured_qubits):
                loss += (l[qubit] - p[qubit]) ** 2

        return loss / len(labels)

    def Cost_function(self, theta, data=None, labels=None, nshots=10000):
        """
        Args:
            theta: list or numpy.array with the biases and the angles to be used in the circuit.
            data: numpy.array data[page][word]  (this is an array of kets).
            labels: list or numpy.array with the labels of the quantum states to be classified.
            nshots: int number of runs of the circuit during the sampling process (default=10000).
        Returns:
            numpy.float32 with the value of the square-loss function.
        """
        circ = self.Classifier_circuit(theta)

        Bias = np.array(theta[0 : self.measured_qubits])
        predictions = np.zeros(shape=(len(data), self.measured_qubits))

        for i, text in enumerate(data):
            predictions[i] = self.Predictions(circ, Bias, text, nshots)

        s = self.square_loss(labels, predictions)

        return s

    def minimize(
        self, init_theta, data=None, labels=None, nshots=10000, method="Powell"
    ):
        """
        Args:
            init_theta: list or numpy.array with the angles to be used in the circuit.
            data: the training data to be used in the minimization.
            labels: the corresponding ground truth for the training data.
            nshots: int number of runs of the circuit during the sampling process (default=10000).
            method: str 'classical optimizer for the minimization'. All methods from qibo.optimizers.optimize are suported (default='Powell').
        Returns:
            numpy.float64 with value of the minimum found, numpy.ndarray with the optimal angles.
        """
        from qibo.optimizers import optimize

        loss, optimal_angles, result = optimize(
            self.Cost_function, init_theta, args=(data, labels, nshots), method=method
        )

        self._optimal_angles = optimal_angles

        self.set_circuit_params(optimal_angles[self.measured_qubits :])
        return loss, optimal_angles

    def Accuracy(self, labels, predictions, sign=True, tolerance=1e-2):
        """
        Args:
            labels: numpy.array with the labels of the quantum states to be classified.
            predictions: numpy.array with the predictions for the quantum states classified.
            sign: if True, labels = np.sign(labels) and predictions = np.sign(predictions) (default=True).
            tolerance: float tolerance level to consider a prediction correct (default=1e-2).
        Returns:
            float with the proportion of states classified successfully.
        """
        if sign:
            labels = [np.sign(label) for label in labels]
            predictions = [np.sign(prediction) for prediction in predictions]

        accur = 0
        for l, p in zip(labels, predictions):
            if np.allclose(l, p, rtol=0.0, atol=tolerance):
                accur += 1

        accur = accur / len(labels)

        return accur

    def predict(self, init_state, nshots=10000):
        """
        This function is used to produce predictions on new input state after the model is trained. Currently it only takes in one input data.

        Args:
            init_state: the input state to be predicted.
            nshots (default=10000): number of shots.
        Returns:
            numpy.array() with predictions for each qubit, for the initial state.
        """
        return self.Predictions(
            self._circuit, self._optimal_angles, init_state, nshots=nshots
        )



from typing import Optional

import networkx as nx

from qibo import gates
from qibo.backends import Backend, _check_backend
from qibo.config import raise_error
from qibo.models import Circuit
from qibo.transpiler.abstract import Optimizer


class Preprocessing(Optimizer):
    """Pad the circuit with unused qubits to match the number of physical qubits.

    Args:
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
    """

    def __init__(self, connectivity: Optional[nx.Graph] = None):
        self.connectivity = connectivity

    def __call__(self, circuit: Circuit) -> Circuit:
        if not all(qubit in self.connectivity.nodes for qubit in circuit.wire_names):
            raise_error(
                ValueError,
                "Some wire_names in the circuit are not in the connectivity graph.",
            )

        physical_qubits = self.connectivity.number_of_nodes()
        logical_qubits = circuit.nqubits
        if logical_qubits > physical_qubits:
            raise_error(
                ValueError,
                f"The number of qubits in the circuit ({logical_qubits}) "
                + f"can't be greater than the number of physical qubits ({physical_qubits}).",
            )
        if logical_qubits == physical_qubits:
            return circuit
        new_wire_names = circuit.wire_names + list(
            self.connectivity.nodes - circuit.wire_names
        )
        new_circuit = Circuit(nqubits=physical_qubits, wire_names=new_wire_names)
        for gate in circuit.queue:
            new_circuit.add(gate)
        return new_circuit


class Rearrange(Optimizer):
    """Rearranges gates using qibo's fusion algorithm.
    May reduce number of SWAPs when fixing for connectivity
    but this has not been tested.

    Args:
        max_qubits (int, optional): Maximum number of qubits to fuse.
            Defaults to :math:`1`.
    """

    def __init__(self, max_qubits: int = 1):
        self.max_qubits = max_qubits

    def __call__(self, circuit: Circuit, backend: Optional[Backend] = None):
        backend = _check_backend(backend)
        fused_circuit = circuit.fuse(max_qubits=self.max_qubits)
        new = circuit.__class__(nqubits=circuit.nqubits, wire_names=circuit.wire_names)
        for fgate in fused_circuit.queue:
            if isinstance(fgate, gates.FusedGate):
                new.add(gates.Unitary(fgate.matrix(backend), *fgate.qubits))
            else:
                new.add(fgate)
        return new



import random
from copy import deepcopy
from typing import Optional, Union

import networkx as nx
import numpy as np

from qibo import gates
from qibo.config import raise_error
from qibo.models import Circuit
from qibo.transpiler._exceptions import ConnectivityError
from qibo.transpiler.abstract import Router
from qibo.transpiler.asserts import assert_placement
from qibo.transpiler.blocks import Block, CircuitBlocks


class StarConnectivityRouter(Router):
    """Transforms an arbitrary circuit to one that can be executed on hardware.

    This transpiler produces a circuit that respects the following connectivity:

             q
             |
        q -- q -- q
             |
             q

    by adding SWAP gates when needed.

    Args:
        connectivity (:class:`networkx.Graph`): Star connectivity graph.
    """

    def __init__(self, connectivity: Optional[nx.Graph] = None):
        self.connectivity = connectivity
        self.middle_qubit = None

    def __call__(self, circuit: Circuit):
        """Apply the transpiler transformation on a given circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): The original Qibo circuit to transform.
                Only single qubit gates and two qubits gates are supported by the router.
        """
        self._check_star_connectivity()
        assert_placement(circuit, self.connectivity)

        middle_qubit_idx = circuit.wire_names.index(self.middle_qubit)
        nqubits = circuit.nqubits
        new = Circuit(nqubits=nqubits, wire_names=circuit.wire_names)
        l2p = list(range(nqubits))

        for i, gate in enumerate(circuit.queue):
            routed_qubits = [l2p[q] for q in gate.qubits]

            if isinstance(gate, gates.M):
                new_gate = gates.M(*routed_qubits, **gate.init_kwargs)
                new_gate.result = gate.result
                new.add(new_gate)
                continue

            if len(routed_qubits) > 2:
                raise_error(
                    ConnectivityError,
                    "Gates targeting more than two qubits are not supported.",
                )

            if len(routed_qubits) == 2 and middle_qubit_idx not in routed_qubits:
                # find which qubit should be moved
                new_middle = _find_connected_qubit(
                    routed_qubits,
                    circuit.queue[i + 1 :],
                    error=ConnectivityError,
                    mapping=l2p,
                )

                new.add(gates.SWAP(new_middle, middle_qubit_idx))
                idx1, idx2 = l2p.index(middle_qubit_idx), l2p.index(new_middle)
                l2p[idx1], l2p[idx2] = l2p[idx2], l2p[idx1]

            routed_qubits = [l2p[q] for q in gate.qubits]

            # add gate to the hardware circuit
            if isinstance(gate, gates.Unitary):
                # gates.Unitary requires matrix as first argument
                matrix = gate.init_args[0]
                new.add(gate.__class__(matrix, *routed_qubits, **gate.init_kwargs))
            else:
                new.add(gate.__class__(*routed_qubits, **gate.init_kwargs))
        return new, {circuit.wire_names[i]: l2p[i] for i in range(nqubits)}

    def _check_star_connectivity(self):
        """Check if the connectivity graph is a star graph."""
        if len(self.connectivity.nodes) != 5:
            raise_error(
                ConnectivityError,
                f"This connectivity graph is not a star graph. Length of nodes provided: {len(self.connectivity.nodes)} != 5.",
            )
        for node in self.connectivity.nodes:
            if self.connectivity.degree(node) == 4:
                self.middle_qubit = node
            elif self.connectivity.degree(node) != 1:
                raise_error(
                    ConnectivityError,
                    "This connectivity graph is not a star graph. There is a node with degree different from 1 and 4.",
                )


def _find_connected_qubit(qubits, queue, error, mapping):
    """Helper method for :meth:`qibo.transpiler.router.StarConnectivityRouter`
    and :meth:`qibo.transpiler.router.StarConnectivityPlacer`.

    Finds which qubit should be mapped to hardware middle qubit
    by looking at the two-qubit gates that follow.
    """
    possible_qubits = {qubits[0], qubits[1]}
    for next_gate in queue:
        if len(next_gate.qubits) > 2:
            raise_error(
                error,
                "Gates targeting more than 2 qubits are not supported",
            )
        if len(next_gate.qubits) == 2:
            possible_qubits &= {
                mapping[next_gate.qubits[0]],
                mapping[next_gate.qubits[1]],
            }

            if len(possible_qubits) == 0:
                return qubits[0]
            if len(possible_qubits) == 1:
                return possible_qubits.pop()

    return qubits[0]


class CircuitMap:
    """Class that stores the circuit and physical-logical mapping during routing.

    Also implements the initial two-qubit block decompositions.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be routed.
        blocks (:class:`qibo.transpiler.blocks.CircuitBlocks`, optional): Circuit block representation.
            If ``None``, the blocks will be computed from the circuit. Defaults to ``None``.
    """

    def __init__(
        self,
        circuit: Optional[Circuit] = None,
        blocks: Optional[CircuitBlocks] = None,
        temp: Optional[bool] = False,
    ):
        self._p2l, self._l2p = [], []

        self._temporary = temp
        if self._temporary:
            return
        if circuit is None:
            raise_error(ValueError, "Circuit must be provided.")
        if blocks is not None:
            self.circuit_blocks = blocks
        else:
            self.circuit_blocks = CircuitBlocks(circuit, index_names=True)

        self.nqubits = circuit.nqubits
        self._routed_blocks = CircuitBlocks(
            Circuit(circuit.nqubits, wire_names=circuit.wire_names)
        )
        self._swaps = 0

        self.wire_names = circuit.wire_names.copy()
        self.physical_to_logical = list(range(self.nqubits))

    @property
    def physical_to_logical(self):
        """Returns the physical to logical qubit mapping."""
        return self._p2l

    @property
    def logical_to_physical(self):
        """Returns the logical to physical qubit mapping."""
        return self._l2p

    @physical_to_logical.setter
    def physical_to_logical(self, p2l_map: list):
        """Sets the physical to logical qubit mapping and updates the logical to physical mapping.

        Args:
            p2l_map (list): Physical to logical mapping.
        """
        self._p2l = p2l_map.copy()
        self._l2p = [0] * len(self._p2l)
        for i, l in enumerate(self._p2l):
            self._l2p[l] = i

    @logical_to_physical.setter
    def logical_to_physical(self, l2p_map: list):
        """Sets the logical to physical qubit mapping and updates the physical to logical mapping.

        Args:
            l2p_map (list): Logical to physical mapping.
        """
        self._l2p = l2p_map.copy()
        self._p2l = [0] * len(self._l2p)
        for i, p in enumerate(self._l2p):
            self._p2l[p] = i

    def _update_mappings_swap(self, logical_swap: tuple, physical_swap: tuple):
        """Updates the qubit mappings after applying a SWAP gate.

        Args:
            logical_swap (tuple[int]): The indices of the logical qubits to be swapped.
            physical_swap (tuple[int]): The indices of the corresponding physical qubits to be swapped.
        """
        self._p2l[physical_swap[0]], self._p2l[physical_swap[1]] = (
            logical_swap[1],
            logical_swap[0],
        )
        self._l2p[logical_swap[0]], self._l2p[logical_swap[1]] = (
            physical_swap[1],
            physical_swap[0],
        )

    def blocks_logical_qubits_pairs(self):
        """Returns a list containing the logical qubit pairs of each block."""
        return [block.qubits for block in self.circuit_blocks()]

    def execute_block(self, block: Block):
        """Executes a block by removing it from the circuit representation
        and adding it to the routed circuit.

        Method works in-place.

        Args:
            block (:class:`qibo.transpiler.blocks.Block`): Block to be removed.
        """
        self._routed_blocks.add_block(block.on_qubits(self.get_physical_qubits(block)))
        self.circuit_blocks.remove_block(block)

    def routed_circuit(self, circuit_kwargs: Optional[dict] = None):
        """Returns the routed circuit.

        Args:
            circuit_kwargs (dict): Original circuit init_kwargs.

        Returns:
            :class:`qibo.models.circuit.Circuit`: Routed circuit.
        """
        return self._routed_blocks.circuit(circuit_kwargs=circuit_kwargs)

    def final_layout(self):
        """Returns the final {logical: physical} qubits mapping."""

        return {self.wire_names[i]: self._l2p[i] for i in range(self.nqubits)}

    def update(self, logical_swap: tuple):
        """Updates the qubit mapping after applying a ``SWAP``

        Adds the :class:`qibo.gates.gates.SWAP` gate to the routed blocks.
        Method works in-place.

        Args:
            swap (tuple): Tuple containing the logical qubits to be swapped.
        """

        physical_swap = self.logical_pair_to_physical(logical_swap)
        if not self._temporary:
            self._routed_blocks.add_block(
                Block(qubits=physical_swap, gates=[gates.SWAP(*physical_swap)])
            )
            self._swaps += 1

        self._update_mappings_swap(logical_swap, physical_swap)

    def undo(self):
        """Undo the last swap. Method works in-place."""
        last_swap_block = self._routed_blocks.return_last_block()
        physical_swap = last_swap_block.qubits
        logical_swap = self._p2l[physical_swap[0]], self._p2l[physical_swap[1]]
        self._routed_blocks.remove_block(last_swap_block)
        self._swaps -= 1

        self._update_mappings_swap(logical_swap, physical_swap)

    def get_physical_qubits(self, block: Union[int, Block]):
        """Returns the physical qubits where a block is acting on.

        Args:
            block (int or :class:`qibo.transpiler.blocks.Block`): Block to be analysed.

        Returns:
            tuple: Physical qubit numbers where a block is acting on.
        """
        if isinstance(block, int):
            block = self.circuit_blocks.search_by_index(block)

        return tuple(self._l2p[q] for q in block.qubits)

    def logical_pair_to_physical(self, logical_qubits: tuple):
        """Returns the physical qubits associated to the logical qubit pair.

        Args:
            logical_qubits (tuple): Logical qubit pair.

        Returns:
            tuple: Physical qubit numbers associated to the logical qubit pair.
        """
        return self._l2p[logical_qubits[0]], self._l2p[logical_qubits[1]]


class ShortestPaths(Router):
    """A class to perform initial qubit mapping and connectivity matching.

    Args:
        connectivity (:class:`networkx.Graph`): Chip connectivity.
        seed (int, optional): Seed for the random number generator.
            If ``None``, defaults to :math:`42`. Defaults to ``None``.
    """

    def __init__(
        self, connectivity: Optional[nx.Graph] = None, seed: Optional[int] = None
    ):
        self.connectivity = connectivity
        self._front_layer = None
        self.circuit_map = None
        self._dag = None
        self._final_measurements = None
        self._node_mapping_inv = None
        if seed is None:
            seed = 42
        random.seed(seed)

    @property
    def added_swaps(self):
        """Returns the number of SWAP gates added to the circuit during routing."""
        return self.circuit_map._swaps

    def __call__(self, circuit: Circuit):
        """Circuit connectivity matching.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be matched to hardware connectivity.

        Returns:
            (:class:`qibo.models.circuit.Circuit`, dict): Routed circuit and final {logical: physical} qubit mapping.
        """
        assert_placement(circuit, self.connectivity)
        self._preprocessing(circuit=circuit)
        while self._dag.number_of_nodes() != 0:
            execute_block_list = self._check_execution()
            if execute_block_list is not None:
                self._execute_blocks(execute_block_list)
            else:
                self._find_new_mapping()

        circuit_kwargs = circuit.init_kwargs
        routed_circuit = self.circuit_map.routed_circuit(circuit_kwargs=circuit_kwargs)
        if self._final_measurements is not None:
            routed_circuit = self._append_final_measurements(
                routed_circuit=routed_circuit
            )

        return routed_circuit, self.circuit_map.final_layout()

    def _find_new_mapping(self):
        """Find new qubit mapping. Mapping is found by looking for the shortest path.

        Method works in-place.
        """
        candidates_evaluation = []
        for candidate in self._candidates():
            cost = self._compute_cost(candidate)
            candidates_evaluation.append((candidate, cost))
        best_cost = min(candidate[1] for candidate in candidates_evaluation)
        best_candidates = [
            candidate[0]
            for candidate in candidates_evaluation
            if candidate[1] == best_cost
        ]
        best_candidate = random.choice(best_candidates)
        self._add_swaps(best_candidate, self.circuit_map)

    def _candidates(self):
        """Returns all possible shortest paths in a ``list`` that contains
        the new mapping and a second ``list`` containing the path meeting point.
        """
        target_qubits = self.circuit_map.get_physical_qubits(self._front_layer[0])
        path_list = list(
            nx.all_shortest_paths(
                self.connectivity, source=target_qubits[0], target=target_qubits[1]
            )
        )
        all_candidates = []
        for path in path_list:
            for meeting_point in range(len(path) - 1):
                all_candidates.append((path, meeting_point))

        return all_candidates

    @staticmethod
    def _add_swaps(candidate: tuple, circuitmap: CircuitMap):
        """Adds swaps to the circuit to move qubits.

        Method works in-place.

        Args:
            candidate (tuple): Contains path to move qubits and qubit meeting point in the path.
            circuitmap (CircuitMap): Representation of the circuit.
        """
        path = candidate[0]
        meeting_point = candidate[1]
        forward = path[0 : meeting_point + 1]
        backward = list(reversed(path[meeting_point + 1 :]))
        for f in forward[1:]:
            circuitmap.update(
                (
                    circuitmap.physical_to_logical[f],
                    circuitmap.physical_to_logical[forward[0]],
                )
            )
        for b in backward[1:]:
            circuitmap.update(
                (
                    circuitmap.physical_to_logical[b],
                    circuitmap.physical_to_logical[backward[0]],
                )
            )

    def _compute_cost(self, candidate: tuple):
        """Greedy algorithm that decides which path to take and how qubits should be walked.

        The cost is computed as minus the number of successive gates that can be executed.

        Args:
            candidate (tuple): Contains path to move qubits and qubit meeting point in the path.

        Returns:
            (list, int): Best path to move qubits and qubit meeting point in the path.
        """
        temporary_circuit = CircuitMap(
            circuit=Circuit(self.circuit_map.nqubits),
            blocks=deepcopy(self.circuit_map.circuit_blocks),
        )

        temporary_circuit.physical_to_logical = self.circuit_map.physical_to_logical
        self._add_swaps(candidate, temporary_circuit)
        temporary_dag = deepcopy(self._dag)
        successive_executed_gates = 0
        while temporary_dag.number_of_nodes() != 0:
            for layer, nodes in enumerate(nx.topological_generations(temporary_dag)):
                for node in nodes:
                    temporary_dag.nodes[node]["layer"] = layer
            temporary_front_layer = [
                node[0] for node in temporary_dag.nodes(data="layer") if node[1] == 0
            ]
            all_executed = True
            for block in temporary_front_layer:
                if (
                    temporary_circuit.get_physical_qubits(block)
                    in self.connectivity.edges
                    or not temporary_circuit.circuit_blocks.search_by_index(
                        block
                    ).entangled
                ):
                    successive_executed_gates += 1
                    temporary_circuit.execute_block(
                        temporary_circuit.circuit_blocks.search_by_index(block)
                    )
                    temporary_dag.remove_node(block)
                else:
                    all_executed = False
            if not all_executed:
                break

        return -successive_executed_gates

    def _check_execution(self):
        """Check if some blocks in the front layer can be executed in the current configuration.

        Returns:
            (list): Executable blocks if there are, ``None`` otherwise.
        """
        executable_blocks = []
        for block in self._front_layer:
            if (
                self.circuit_map.get_physical_qubits(block) in self.connectivity.edges
                or not self.circuit_map.circuit_blocks.search_by_index(block).entangled
            ):
                executable_blocks.append(block)
        if len(executable_blocks) == 0:
            return None

        return executable_blocks

    def _execute_blocks(self, blocklist: list):
        """Executes a list of blocks:
            -Remove the correspondent nodes from the dag and circuit representation.
            -Add the executed blocks to the routed circuit.
            -Update the dag layers and front layer.

        Method works in-place.

        Args:
            blocklist (list): List of blocks.
        """
        for block_id in blocklist:
            block = self.circuit_map.circuit_blocks.search_by_index(block_id)
            self.circuit_map.execute_block(block)
            self._dag.remove_node(block_id)
        self._update_front_layer()

    def _update_front_layer(self):
        """Updates the front layer of the dag.

        Method works in-place.
        """
        for layer, nodes in enumerate(nx.topological_generations(self._dag)):
            for node in nodes:
                self._dag.nodes[node]["layer"] = layer
        self._front_layer = [
            node[0] for node in self._dag.nodes(data="layer") if node[1] == 0
        ]

    def _preprocessing(self, circuit: Circuit):
        """The following objects will be initialised:
            - circuit: class to represent circuit and to perform {logical: physical} qubit mapping.
            - _final_measurements: measurement gates at the end of the circuit.
            - _front_layer: list containing the blocks to be executed.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be preprocessed.
        """
        self.connectivity = nx.relabel_nodes(
            self.connectivity, {v: i for i, v in enumerate(circuit.wire_names)}
        )
        copied_circuit = circuit.copy(deep=True)
        self._final_measurements = self._detach_final_measurements(copied_circuit)
        self.circuit_map = CircuitMap(copied_circuit)
        self._dag = _create_dag(self.circuit_map.blocks_logical_qubits_pairs())
        self._update_front_layer()

    def _detach_final_measurements(self, circuit: Circuit):
        """Detaches measurement gates at the end of the circuit for separate handling.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuits to be processed.

        Returns:
            (NoneType or list): List of measurements. If no measurements, returns ``None``.
        """
        final_measurements = []
        for gate in circuit.queue[::-1]:
            if isinstance(gate, gates.M):
                final_measurements.append(gate)
                circuit.queue.remove(gate)
            else:
                break
        if not final_measurements:
            return None

        return final_measurements[::-1]

    def _append_final_measurements(self, routed_circuit: Circuit):
        """Appends the final measurment gates on the correct qubits
        conserving the measurement register."""
        for measurement in self._final_measurements:
            original_qubits = measurement.qubits
            routed_qubits = list(
                self.circuit_map.logical_to_physical[qubit] for qubit in original_qubits
            )
            routed_circuit.add(
                measurement.on_qubits(dict(zip(original_qubits, routed_qubits)))
            )

        return routed_circuit


class Sabre(Router):
    """Routing algorithm proposed in Ref [1].

    Args:
        connectivity (:class:`networkx.Graph`): Hardware chip connectivity.
        lookahead (int, optional): Lookahead factor, how many dag layers will be considered
            in computing the cost. Defaults to :math:`2`.
        decay_lookahead (float, optional): Value in interval :math:`[0, 1]`.
            How the weight of the distance in the dag layers decays in computing the cost.
            Defaults to :math:`0.6`.
        delta (float, optional): Defines the number of SWAPs vs depth trade-off by deciding
            how the algorithm tends to select non-overlapping SWAPs.
            Defaults to math:`10^{-3}`.
        seed (int, optional): Seed for the candidate random choice as tiebraker.
            Defaults to ``None``.
        swap_threshold (float, optional): Limits the number of added SWAPs in every routing iteration.
            This threshold is multiplied by the length of the longest path in the circuit connectivity.
            If the number of added SWAPs exceeds the threshold before a gate is routed,
            shortestpath routing is applied.
            Defaults to :math:`1.5`.

    References:
        1. G. Li, Y. Ding, and Y. Xie,
        *Tackling the Qubit Mapping Problem for NISQ-Era Quantum Devices*.
        `arXiv:1809.02573 [cs.ET] <https://arxiv.org/abs/1809.02573>`_.
    """

    def __init__(
        self,
        connectivity: Optional[nx.Graph] = None,
        lookahead: int = 2,
        decay_lookahead: float = 0.6,
        delta: float = 0.001,
        swap_threshold: float = 1.5,
        seed: Optional[int] = None,
    ):
        self.connectivity = connectivity
        self.lookahead = lookahead
        self.decay = decay_lookahead
        self.delta = delta
        self.swap_threshold = swap_threshold
        self._delta_register = None
        self._dist_matrix = None
        self._dag = None
        self._front_layer = None
        self.circuit_map = None
        self._memory_map = None
        self._final_measurements = None
        self._temp_added_swaps = []
        random.seed(seed)

    def __call__(self, circuit: Circuit):
        """Route the circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be routed.

        Returns:
            (:class:`qibo.models.circuit.Circuit`, dict): Routed circuit and final {logical: physical} qubit mapping.
        """
        assert_placement(circuit, self.connectivity)
        self._preprocessing(circuit=circuit)
        longest_path = np.max(self._dist_matrix)

        while self._dag.number_of_nodes() != 0:
            execute_block_list = self._check_execution()
            if execute_block_list is not None:
                self._execute_blocks(execute_block_list)
            else:
                self._find_new_mapping()

            # If the number of added swaps is too high, the algorithm is stuck.
            # Reset the circuit to the last saved state and make the nearest gate executable by manually adding SWAPs.
            if (
                len(self._temp_added_swaps) > self.swap_threshold * longest_path
            ):  # threshold is arbitrary
                while self._temp_added_swaps:
                    self._temp_added_swaps.pop()
                    self.circuit_map.undo()
                self._temp_added_swaps = []
                self._shortest_path_routing()

        circuit_kwargs = circuit.init_kwargs
        routed_circuit = self.circuit_map.routed_circuit(circuit_kwargs=circuit_kwargs)
        if self._final_measurements is not None:
            routed_circuit = self._append_final_measurements(
                routed_circuit=routed_circuit
            )

        return routed_circuit, self.circuit_map.final_layout()

    @property
    def added_swaps(self):
        """Returns the number of SWAP gates added to the circuit during routing."""
        return self.circuit_map._swaps

    def _preprocessing(self, circuit: Circuit):
        """The following objects will be initialised:
            - circuit: class to represent circuit and to perform {logical: physical} qubit mapping.
            - _final_measurements: measurement gates at the end of the circuit.
            - _dist_matrix: matrix reporting the shortest path lengh between all node pairs.
            - _dag: direct acyclic graph of the circuit based on commutativity.
            - _memory_map: list to remember previous SWAP moves.
            - _front_layer: list containing the blocks to be executed.
            - _delta_register: list containing the special weigh added to qubits
                to prevent overlapping swaps.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be preprocessed.
        """

        self.connectivity = nx.relabel_nodes(
            self.connectivity, {v: i for i, v in enumerate(circuit.wire_names)}
        )

        copied_circuit = circuit.copy(deep=True)
        self._final_measurements = self._detach_final_measurements(copied_circuit)
        self.circuit_map = CircuitMap(copied_circuit)
        self._dist_matrix = nx.floyd_warshall_numpy(self.connectivity)
        self._dag = _create_dag(self.circuit_map.blocks_logical_qubits_pairs())
        self._memory_map = []
        self._update_dag_layers()
        self._update_front_layer()
        self._delta_register = [1.0 for _ in range(circuit.nqubits)]

    def _detach_final_measurements(self, circuit: Circuit):
        """Detach measurement gates at the end of the circuit for separate handling."""
        final_measurements = []
        for gate in circuit.queue[::-1]:
            if isinstance(gate, gates.M):
                final_measurements.append(gate)
                circuit.queue.remove(gate)
            else:
                break
        if not final_measurements:
            return None
        return final_measurements[::-1]

    def _append_final_measurements(self, routed_circuit: Circuit):
        """Appends final measurment gates on the correct qubits conserving the measurement register.

        Args:
            routed_circuit (:class:`qibo.models.circuit.Circuit`): Original circuit.

        Returns:
            (:class:`qibo.models.circuit.Circuit`) Routed circuit.
        """
        for measurement in self._final_measurements:
            original_qubits = measurement.qubits
            routed_qubits = list(
                self.circuit_map.logical_to_physical[qubit] for qubit in original_qubits
            )
            routed_circuit.add(
                measurement.on_qubits(dict(zip(original_qubits, routed_qubits)))
            )
        return routed_circuit

    def _update_dag_layers(self):
        """Update dag layers and put them in topological order.

        Method works in-place.
        """
        for layer, nodes in enumerate(nx.topological_generations(self._dag)):
            for node in nodes:
                self._dag.nodes[node]["layer"] = layer

    def _update_front_layer(self):
        """Update the front layer of the dag.

        Method works in-place.
        """
        self._front_layer = self._get_dag_layer(0)

    def _get_dag_layer(self, n_layer, qubits=False):
        """Return the :math:`n`-topological layer of the dag.

        Args:
            n_layer (int): Layer number.
            qubits (bool, optional): If ``True``, return the target qubits of the blocks in the layer.
                If ``False``, return the block numbers. Defaults to ``False``.

        Returns:
            (list): List of block numbers or target qubits.
        """

        if qubits:
            return [
                node[1]["qubits"]
                for node in self._dag.nodes(data=True)
                if node[1]["layer"] == n_layer
            ]

        return [node[0] for node in self._dag.nodes(data="layer") if node[1] == n_layer]

    def _find_new_mapping(self):
        """Find the new best mapping by adding one swap."""
        candidates_evaluation = {}

        self._memory_map.append(self.circuit_map.physical_to_logical.copy())
        for candidate in self._swap_candidates():
            candidates_evaluation[candidate] = self._compute_cost(candidate)

        best_cost = min(candidates_evaluation.values())
        best_candidates = [
            key for key, value in candidates_evaluation.items() if value == best_cost
        ]
        best_candidate = random.choice(best_candidates)

        for qubit in self.circuit_map.logical_pair_to_physical(best_candidate):
            self._delta_register[qubit] += self.delta
        self.circuit_map.update(best_candidate)
        self._temp_added_swaps.append(best_candidate)

    def _compute_cost(self, candidate: int):
        """Compute the cost associated to a possible SWAP candidate."""

        temporary_circuit = CircuitMap(temp=True)
        temporary_circuit.physical_to_logical = self.circuit_map.physical_to_logical
        temporary_circuit.update(candidate)

        if temporary_circuit.physical_to_logical in self._memory_map:
            return float("inf")

        tot_distance = 0.0
        weight = 1.0
        for layer in range(self.lookahead + 1):
            layer_gates = self._get_dag_layer(layer, qubits=True)
            avg_layer_distance = 0.0
            for lq_pair in layer_gates:
                qubits = temporary_circuit.logical_pair_to_physical(lq_pair)
                avg_layer_distance += (
                    max(self._delta_register[i] for i in qubits)
                    * (self._dist_matrix[qubits[0], qubits[1]] - 1.0)
                    / len(layer_gates)
                )
            tot_distance += weight * avg_layer_distance
            weight *= self.decay

        return tot_distance

    def _swap_candidates(self):
        """Returns a list of possible candidate SWAPs to be applied on logical qubits directly.

        The possible candidates are the ones sharing at least one qubit
        with a block in the front layer.

        Returns:
            (list): List of candidates.
        """
        candidates = []
        for block in self._front_layer:
            for qubit in self.circuit_map.get_physical_qubits(block):
                for connected in self.connectivity.neighbors(qubit):
                    candidate = tuple(
                        sorted(
                            (
                                self.circuit_map.physical_to_logical[qubit],
                                self.circuit_map.physical_to_logical[connected],
                            )
                        )
                    )
                    if candidate not in candidates:
                        candidates.append(candidate)

        return candidates

    def _check_execution(self):
        """Check if some blocks in the front layer can be executed in the current configuration.

        Returns:
            (list): Executable blocks if there are, ``None`` otherwise.
        """
        executable_blocks = []
        for block in self._front_layer:
            if (
                self.circuit_map.get_physical_qubits(block) in self.connectivity.edges
                or not self.circuit_map.circuit_blocks.search_by_index(block).entangled
            ):
                executable_blocks.append(block)

        if len(executable_blocks) == 0:
            return None

        return executable_blocks

    def _execute_blocks(self, blocklist: list):
        """Executes a list of blocks:
        -Remove the correspondent nodes from the dag and circuit representation.
        -Add the executed blocks to the routed circuit.
        -Update the dag layers and front layer.
        -Reset the mapping memory.

        Method works in-place.

        Args:
            blocklist (list): List of blocks.
        """
        for block_id in blocklist:
            block = self.circuit_map.circuit_blocks.search_by_index(block_id)
            self.circuit_map.execute_block(block)
            self._dag.remove_node(block_id)
        self._update_dag_layers()
        self._update_front_layer()
        self._memory_map = []
        self._delta_register = [1.0 for _ in self._delta_register]
        self._temp_added_swaps = []

    def _shortest_path_routing(self):
        """Route a gate in the front layer using the shortest path. This method is executed when the standard SABRE fails to find an optimized solution.

        Method works in-place.
        """

        min_distance = float("inf")
        shortest_path_qubits = None

        for block in self._front_layer:
            q1, q2 = self.circuit_map.get_physical_qubits(block)
            distance = self._dist_matrix[q1, q2]

            if distance < min_distance:
                min_distance = distance
                shortest_path_qubits = [q1, q2]

        shortest_path = nx.bidirectional_shortest_path(
            self.connectivity, shortest_path_qubits[0], shortest_path_qubits[1]
        )

        # move q1
        q1 = self.circuit_map.physical_to_logical[shortest_path[0]]
        for q2 in shortest_path[1:-1]:
            self.circuit_map.update((q1, self.circuit_map.physical_to_logical[q2]))


def _create_dag(gates_qubits_pairs: list):
    """Helper method for :meth:`qibo.transpiler.router.Sabre`.

    Create direct acyclic graph (dag) of the circuit based on two qubit gates
    commutativity relations.

    Args:
        gates_qubits_pairs (list): List of qubits tuples where gates/blocks acts.

    Returns:
        (:class:`networkx.DiGraph`): Adjoint of the circuit.
    """
    dag = nx.DiGraph()
    dag.add_nodes_from(range(len(gates_qubits_pairs)))

    for i, _ in enumerate(gates_qubits_pairs):
        dag.nodes[i]["qubits"] = gates_qubits_pairs[i]

    # Find all successors
    connectivity_list = []
    for idx, gate in enumerate(gates_qubits_pairs):
        saturated_qubits = []
        for next_idx, next_gate in enumerate(gates_qubits_pairs[idx + 1 :]):
            for qubit in gate:
                if (qubit in next_gate) and (qubit not in saturated_qubits):
                    saturated_qubits.append(qubit)
                    connectivity_list.append((idx, next_idx + idx + 1))
            if len(saturated_qubits) >= 2:
                break
    dag.add_edges_from(connectivity_list)

    return _remove_redundant_connections(dag)


def _remove_redundant_connections(dag: nx.DiGraph):
    """Helper method for :func:`qibo.transpiler.router._create_dag`.

    Remove redundant connection from a DAG using transitive reduction.

    Args:
        dag (:class:`networkx.DiGraph`): DAG to be reduced.

    Returns:
        (:class:`networkx.DiGraph`): Reduced DAG.
    """
    new_dag = nx.DiGraph()
    new_dag.add_nodes_from(dag.nodes(data=True))
    transitive_reduction = nx.transitive_reduction(dag)
    new_dag.add_edges_from(transitive_reduction.edges)

    return new_dag



import abc
from typing import Optional, Union

from qibo.config import raise_error


class Backend(abc.ABC):
    def __init__(self):
        super().__init__()
        self.name = "backend"
        self.platform = None

        self.dtype = "complex128"
        self.matrices = None

        self.device = "/CPU:0"
        self.nthreads = 1
        self.supports_multigpu = False
        self.oom_error = MemoryError

    def __reduce__(self):
        """Allow pickling backend objects that have references to modules."""
        return self.__class__, tuple()

    def __repr__(self):
        if self.platform is None:
            return self.name

        return f"{self.name} ({self.platform})"

    @property
    @abc.abstractmethod
    def qubits(self) -> Optional[list[Union[int, str]]]:  # pragma: no cover
        """Return the qubit names of the backend. If :class:`SimulationBackend`, return None."""
        raise_error(NotImplementedError)

    @property
    @abc.abstractmethod
    def connectivity(
        self,
    ) -> Optional[list[tuple[Union[int, str], Union[int, str]]]]:  # pragma: no cover
        """Return the available qubit pairs of the backend. If :class:`SimulationBackend`, return None."""
        raise_error(NotImplementedError)

    @property
    @abc.abstractmethod
    def natives(self) -> Optional[list[str]]:  # pragma: no cover
        """Return the native gates of the backend. If :class:`SimulationBackend`, return None."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_dtype(self, dtype: str):  # pragma: no cover
        """Set data type of arrays created using the backend.

        .. note::
            The data types ``float32`` and ``float64`` are intended to be used when the circuits
            to be simulated only contain gates with real-valued matrix representations.
            Using one of the aforementioned data types with circuits that contain complex-valued
            matrices will raise a casting error.

        .. note::
            List of gates that always admit a real-valued matrix representation:
            :class:`qibo.gates.I`, :class:`qibo.gates.X`, :class:`qibo.gates.Z`,
            :class:`qibo.gates.H`, :class:`qibo.gates.Align`, :class:`qibo.gates.RY`,
            :class:`qibo.gates.CNOT`, :class:`qibo.gates.CZ`, :class:`qibo.gates.CRY`,
            :class:`qibo.gates.SWAP`, :class:`qibo.gates.FSWAP`, :class:`qibo.gates.GIVENS`,
            :class:`qibo.gates.RBS`, :class:`qibo.gates.TOFFOLI`, :class:`qibo.gates.CCZ`,
            and :class:`qibo.gates.FanOut`.

        .. note::
            The following parametrized gates can have real-valued matrix representations
            depending on the values of their parameters:
            :class:`qibo.gates.RX`, :class:`qibo.gates.RZ`, :class:`qibo.gates.U1`,
            :class:`qibo.gates.U2`, :class:`qibo.gates.U3`, :class:`qibo.gates.CRX`,
            :class:`qibo.gates.CRZ`, :class:`qibo.gates.CU1`, :class:`qibo.gates.CU2`,
            :class:`qibo.gates.CU3`, :class:`qibo.gates.fSim`, :class:`qibo.gates.GeneralizedfSim`,
            :class:`qibo.gates.RXX`, :class:`qibo.gates.RYY`, :class:`qibo.gates.RZZ`,
            :class:`qibo.gates.RZX`, and :class:`qibo.gates.GeneralizedRBS`.

        Args:
            dtype (str): the options are the following: ``complex128``, ``complex64``,
                ``float64``, and ``float32``.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_device(self, device: str):  # pragma: no cover
        """Set simulation device.

        Args:
            device (str): Device such as '/CPU:0', '/GPU:0', etc.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_threads(self, nthreads: int):  # pragma: no cover
        """Set number of threads for CPU simulation.

        Args:
            nthreads (int): Number of threads.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def cast(self, x, dtype=None, copy: bool = False):  # pragma: no cover
        """Cast an object as the array type of the current backend.

        Args:
            x: Object to cast to array.
            dtype (str or type, optional): data type of ``x`` after casting.
                Options are ``"complex128"``, ``"complex64"``, ``"float64"``,
                or ``"float32"``. If ``None``, defaults to ``Backend.dtype``.
                Defaults to ``None``.
            copy (bool, optional): If ``True`` a copy of the object is created in memory.
                Defaults to ``False``.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def is_sparse(self, x):  # pragma: no cover
        """Determine if a given array is a sparse tensor."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def to_numpy(self, x):  # pragma: no cover
        """Cast a given array to numpy."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def compile(self, func):  # pragma: no cover
        """Compile the given method.

        Available only for the tensorflow backend.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def zero_state(self, nqubits: int):  # pragma: no cover
        """Generate :math:`|000 \\cdots 0 \\rangle` state vector as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def zero_density_matrix(self, nqubits: int):  # pragma: no cover
        """Generate :math:`|000\\cdots0\\rangle\\langle000\\cdots0|` density matrix as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def identity_density_matrix(
        self, nqubits: int, normalize: bool = True
    ):  # pragma: no cover
        """Generate density matrix

        .. math::
            \\rho = \\frac{1}{2^\\text{nqubits}} \\, \\sum_{k=0}^{2^\\text{nqubits} - 1} \\,
                |k \\rangle \\langle k|

        if ``normalize=True``. If ``normalize=False``, returns the unnormalized
        Identity matrix, which is equivalent to :func:`numpy.eye`.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def plus_state(self, nqubits: int):  # pragma: no cover
        """Generate :math:`|+++\\cdots+\\rangle` state vector as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def plus_density_matrix(self, nqubits: int):  # pragma: no cover
        """Generate :math:`|+++\\cdots+\\rangle\\langle+++\\cdots+|` density matrix as an array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix(self, gate: "qibo.gates.abstract.Gate"):  # pragma: no cover
        """Convert a :class:`qibo.gates.Gate` to the corresponding matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix_parametrized(self, gate: "qibo.gates.abstract.Gate"):  # pragma: no cover
        """Equivalent to :meth:`qibo.backends.abstract.Backend.matrix` for parametrized gates."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def matrix_fused(self, gate):  # pragma: no cover
        """Fuse matrices of multiple gates."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate(self, gate, state, nqubits: int):  # pragma: no cover
        """Apply a gate to state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate_density_matrix(self, gate, state, nqubits: int):  # pragma: no cover
        """Apply a gate to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_gate_half_density_matrix(
        self, gate, state, nqubits: int
    ):  # pragma: no cover
        """Apply a gate to one side of the density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_channel(self, channel, state, nqubits: int):  # pragma: no cover
        """Apply a channel to state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def apply_channel_density_matrix(
        self, channel, state, nqubits: int
    ):  # pragma: no cover
        """Apply a channel to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def collapse_state(
        self, state, qubits, shot, nqubits: int, normalize: bool = True
    ):  # pragma: no cover
        """Collapse state vector according to measurement shot."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def collapse_density_matrix(
        self, state, qubits, shot, nqubits: int, normalize: bool = True
    ):  # pragma: no cover
        """Collapse density matrix according to measurement shot."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def reset_error_density_matrix(self, gate, state, nqubits: int):  # pragma: no cover
        """Apply reset error to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def thermal_error_density_matrix(
        self, gate, state, nqubits: int
    ):  # pragma: no cover
        """Apply thermal relaxation error to density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuit(
        self, circuit, initial_state=None, nshots: int = None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit`."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuits(
        self, circuits, initial_states=None, nshots: int = None
    ):  # pragma: no cover
        """Execute multiple :class:`qibo.models.circuit.Circuit` in parallel."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_circuit_repeated(
        self, circuit: "qibo.models.circuit.Circuit", nshots: int, initial_state=None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit` multiple times.

        Useful for noise simulation using state vectors or for simulating gates
        controlled by measurement outcomes.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def execute_distributed_circuit(
        self, circuit, initial_state=None, nshots: int = None
    ):  # pragma: no cover
        """Execute a :class:`qibo.models.circuit.Circuit` using multiple GPUs."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_symbolic(
        self,
        state,
        nqubits: int,
        decimals: int = 5,
        cutoff: float = 1e-10,
        max_terms: int = 20,
    ):  # pragma: no cover
        """Dirac representation of a state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_symbolic_density_matrix(
        self,
        state,
        nqubits: int,
        decimals: int = 5,
        cutoff: float = 1e-10,
        max_terms: int = 20,
    ):  # pragma: no cover
        """Dirac representation of a density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_probabilities(self, state, qubits, nqubits: int):  # pragma: no cover
        """Calculate probabilities given a state vector."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_probabilities_density_matrix(
        self, state, qubits, nqubits: int
    ):  # pragma: no cover
        """Calculate probabilities given a density matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def set_seed(self, seed):  # pragma: no cover
        """Set the seed of the random number generator."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def sample_shots(self, probabilities, nshots: int):  # pragma: no cover
        """Sample measurement shots according to a probability distribution."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def aggregate_shots(self, shots):  # pragma: no cover
        """Collect shots to a single array."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def samples_to_binary(self, samples, nqubits: int):  # pragma: no cover
        """Convert samples from decimal representation to binary."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def samples_to_decimal(self, samples, nqubits: int):  # pragma: no cover
        """Convert samples from binary representation to decimal."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_frequencies(self, samples):  # pragma: no cover
        """Calculate measurement frequencies from shots."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def update_frequencies(
        self, frequencies, probabilities, nsamples: int
    ):  # pragma: no cover
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def sample_frequencies(self, probabilities, nshots: int):  # pragma: no cover
        """Sample measurement frequencies according to a probability distribution."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_vector_norm(
        self, state, order: Union[int, float, str] = 2
    ):  # pragma: no cover
        """Calculate norm of an :math:`1`-dimensional array.

        For specifications on possible values of the parameter ``order``
        for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_norm(
        self, state, order: Union[int, float, str] = "nuc"
    ):  # pragma: no cover
        """Calculate norm of a :math:`2`-dimensional array.

        Default is the ``nuclear`` norm.
        If ``order="nuc"``, it returns the nuclear norm of ``state``,
        assuming ``state`` is Hermitian (also known as trace norm).
        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_overlap(self, state1, state2):  # pragma: no cover
        """Calculate overlap of two state vectors."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_overlap_density_matrix(self, state1, state2):  # pragma: no cover
        """Calculate overlap of two density matrices."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_eigenvalues(
        self, matrix, k: int = 6, hermitian: bool = True
    ):  # pragma: no cover
        """Calculate eigenvalues of a matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_eigenvectors(
        self, matrix, k: int = 6, hermitian: bool = True
    ):  # pragma: no cover
        """Calculate eigenvectors of a matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_expectation_state(
        self, hamiltonian, state, normalize: bool
    ):  # pragma: no cover
        """Calculate expectation value of a state vector given the observable matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_expectation_density_matrix(
        self, hamiltonian, state, normalize: bool
    ):  # pragma: no cover
        """Calculate expectation value of a density matrix given the observable matrix."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_exp(
        self,
        matrix,
        phase: Union[float, int, complex] = 1,
        eigenvectors=None,
        eigenvalues=None,
    ):  # pragma: no cover
        """Calculate the exponential :math:`e^{\\theta \\, A}` of a matrix :math:`A`
        and ``phase`` :math:`\\theta`.

        If the eigenvectors and eigenvalues are given the matrix diagonalization is
        used for exponentiation.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_log(
        self, matrix, base: Union[float, int] = 2, eigenvectors=None, eigenvalues=None
    ):  # pragma: no cover
        """Calculate the logarithm :math:`\\log_{b}(A)` with a ``base`` :math:`b`
        of a matrix :math:`A`.

        If the eigenvectors and eigenvalues are given the matrix diagonalization is
        used for exponentiation.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_power(
        self, matrix, power: Union[float, int], precision_singularity: float = 1e-14
    ):  # pragma: no cover
        """Calculate the (fractional) ``power`` :math:`\\alpha` of ``matrix`` :math:`A`,
        i.e. :math:`A^{\\alpha}`.

        .. note::
            For the ``pytorch`` backend, this method relies on a copy of the original tensor.
            This may break the gradient flow. For the GPU backends (i.e. ``cupy`` and
            ``cuquantum``), this method falls back to CPU whenever ``power`` is not
            an integer.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_matrix_sqrt(
        self, matrix, precision_singularity: float = 1e-14
    ):  # pragma: no cover
        """Calculate the square root of ``matrix`` :math:`A`, i.e. :math:`A^{1/2}`.

        .. note::
            For the ``pytorch`` backend, this method relies on a copy of the original tensor.
            This may break the gradient flow. For the GPU backends (i.e. ``cupy`` and
            ``cuquantum``), this method falls back to CPU.
        """
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_singular_value_decomposition(self, matrix):  # pragma: no cover
        """Calculate the Singular Value Decomposition of ``matrix``."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def calculate_jacobian_matrix(
        self, circuit, parameters, initial_state=None, return_complex: bool = True
    ):  # pragma: no cover
        """Calculate the Jacobian matrix of ``circuit`` with respect to varables ``params``."""
        raise_error(NotImplementedError)

    @abc.abstractmethod
    def assert_allclose(
        self, value, target, rtol: float = 1e-7, atol: float = 0.0
    ):  # pragma: no cover
        raise_error(NotImplementedError)

    def assert_circuitclose(
        self, circuit, target_circuit, rtol: float = 1e-7, atol: float = 0.0
    ):
        value = self.execute_circuit(circuit)._state
        target = self.execute_circuit(target_circuit)._state
        self.assert_allclose(value, target, rtol=rtol, atol=atol)



import numpy as np
from numpy import ndarray
from scipy.linalg import expm

ENGINE = np


def _pauli_basis(
    nqubits: int,
    pauli_0: ndarray,
    pauli_1: ndarray,
    pauli_2: ndarray,
    pauli_3: ndarray,
    normalization: float = 1.0,
) -> ndarray:
    basis_single = ENGINE.vstack((pauli_0, pauli_1, pauli_2, pauli_3)).reshape(4, 2, 2)
    dim = 2**nqubits
    input_indices = [range(3 * i, 3 * (i + 1)) for i in range(nqubits)]
    output_indices = [i for indices in zip(*input_indices) for i in indices]
    operands = [basis_single for _ in range(nqubits)]
    inputs = [item for pair in zip(operands, input_indices) for item in pair]
    return (
        ENGINE.einsum(  # pylint: disable=too-many-function-args
            *inputs, output_indices
        ).reshape(dim**2, dim, dim)
        / normalization
    )


def _post_sparse_pauli_basis_vectorization(
    basis: ndarray, dim: int
) -> tuple[ndarray, ndarray]:
    indices = ENGINE.nonzero(basis)
    basis = basis[indices].reshape(-1, dim)
    indices = indices[1].reshape(-1, dim)
    return basis, indices


_vectorize_pauli_basis = """
def _vectorize_pauli_basis_{order}(
    nqubits: int, pauli_0: ndarray, pauli_1: ndarray, pauli_2: ndarray, pauli_3: ndarray, normalization: float = 1.
) -> ndarray:
    dim = 2**nqubits
    basis = _pauli_basis(nqubits, pauli_0, pauli_1, pauli_2, pauli_3, normalization)
    return _vectorization_{order}(basis, dim)
"""

_vectorize_sparse_pauli_basis = """
def _vectorize_sparse_pauli_basis_{order}(
    nqubits: int, pauli_0: ndarray, pauli_1: ndarray, pauli_2: ndarray, pauli_3: ndarray, normalization: float = 1.
) -> tuple[ndarray, ndarray]:
    dim = 2**nqubits
    basis = _vectorize_pauli_basis_{order}(nqubits, pauli_0, pauli_1, pauli_2, pauli_3, normalization)
    return _post_sparse_pauli_basis_vectorization(basis, dim)
"""

_pauli_to_comp_basis = """
def _pauli_to_comp_basis_sparse_{order}(
        nqubits: int, pauli_0: ndarray, pauli_1: ndarray, pauli_2: ndarray, pauli_3: ndarray, normalization: float = 1.
) -> ndarray:
    unitary = _vectorize_pauli_basis_{order}(nqubits, pauli_0, pauli_1, pauli_2, pauli_3, normalization).T
    nonzero = ENGINE.nonzero(unitary)
    return unitary[nonzero].reshape(unitary.shape[0], -1), nonzero[1]
"""

_super_op_from_haar_measure = """
def _super_op_from_haar_measure_{order}(dims: int) -> ndarray:
    super_op = _random_unitary_haar(dims)
    super_op = _vectorization_{order}(super_op, dims).ravel()
    return ENGINE.outer(super_op, ENGINE.conj(super_op))
"""

_super_op_from_hermitian_measure = """
def _super_op_from_hermitian_measure_{order}(dims: int) -> ndarray:
    super_op = _random_unitary(dims)
    super_op = _vectorization_{order}(super_op, dims).ravel()
    return ENGINE.outer(super_op, ENGINE.conj(super_op))
"""

_to_choi = """
def _to_choi_{order}(channel: ndarray) -> ndarray:
    channel = _vectorization_{order}(channel, channel.shape[-1]).ravel()
    return ENGINE.outer(channel, ENGINE.conj(channel))
"""

_to_liouville = """
def _to_liouville_{order}(channel: ndarray) -> ndarray:
    channel = _to_choi_{order}(channel)
    return _reshuffling(channel, {ax1}, {ax2})
"""

exec(_to_liouville.format(order="row", ax1=1, ax2=2))
exec(_to_liouville.format(order="column", ax1=0, ax2=3))


_to_pauli_liouville = """
def _to_pauli_liouville_{order}(
        channel: ndarray, pauli_0: ndarray, pauli_1: ndarray, pauli_2: ndarray, pauli_3: ndarray, normalization: float = 1.
) -> ndarray:
    nqubits = int(np.log2(channel.shape[0]))
    channel = _to_liouville_{order}(channel)
    unitary = _vectorize_pauli_basis_{order}(
        nqubits, pauli_0, pauli_1, pauli_2, pauli_3, normalization
    )
    return ENGINE.conj(unitary) @ channel @ unitary.T
"""

_choi_to_kraus = """
def _choi_to_kraus_{order}(choi_super_op: ndarray) -> tuple[ndarray, ndarray]:
    U, coefficients, V = ENGINE.linalg.svd(choi_super_op)
    U = U.T
    coefficients = ENGINE.sqrt(coefficients)
    V = ENGINE.conj(V)
    dim = int(np.sqrt(U.shape[-1]))
    coefficients = coefficients.reshape(U.shape[0], 1, 1)
    kraus_left = coefficients * _unvectorization_{order}(U, dim)
    kraus_right = coefficients * _unvectorization_{order}(V, dim)
    kraus_ops = ENGINE.vstack((kraus_left[None,:], kraus_right[None,:]))
    return kraus_ops, coefficients
"""

_choi_to_kraus_cp = """
def _choi_to_kraus_cp_{order}(
    eigenvalues: ndarray, eigenvectors: ndarray, precision: float
) -> tuple[ndarray, ndarray]:
    eigv_gt_tol = ENGINE.abs(eigenvalues) > precision
    coefficients = ENGINE.sqrt(eigenvalues[eigv_gt_tol])
    eigenvectors = eigenvectors[eigv_gt_tol]
    dim = int(np.sqrt(eigenvectors.shape[-1]))
    kraus_ops = coefficients.reshape(-1, 1, 1) * _unvectorization_{order}(
        eigenvectors, dim
    )
    return kraus_ops, coefficients
"""

_kraus_to_choi = """
def _kraus_to_choi_{order}(kraus_ops: ndarray) -> ndarray:
    kraus_ops = _vectorization_{order}(kraus_ops, kraus_ops.shape[-1])
    return kraus_ops.T @ ENGINE.conj(kraus_ops)
"""

for order in ("row", "column", "system"):
    for func in (
        _vectorize_pauli_basis,
        _vectorize_sparse_pauli_basis,
        _pauli_to_comp_basis,
        _super_op_from_haar_measure,
        _super_op_from_hermitian_measure,
        _to_choi,
        _to_pauli_liouville,
        _choi_to_kraus,
        _choi_to_kraus_cp,
        _kraus_to_choi,
    ):
        exec(func.format(order=order))


def _vectorization_row(state: ndarray, dim: int) -> ndarray:
    return ENGINE.reshape(state, (-1, dim**2))


def _vectorization_column(state: ndarray, dim: int) -> ndarray:
    indices = list(range(state.ndim))
    indices[-2:] = indices[-2:][::-1]
    state = ENGINE.transpose(state, indices)
    return ENGINE.reshape(state, (-1, dim**2))


# WARNING: dim is not used, but it is useful to uniform
# the call with the other functions
def _vectorization_system(state: ndarray, dim: int = 0) -> ndarray:
    nqubits = int(np.log2(state.shape[-1]))
    new_axis = [
        0,
    ]
    for qubit in range(nqubits):
        new_axis.extend([qubit + nqubits + 1, qubit + 1])
    state = ENGINE.reshape(state, [-1] + [2] * 2 * nqubits)
    state = ENGINE.transpose(state, new_axis)
    return ENGINE.reshape(state, (-1, 2 ** (2 * nqubits)))


def _unvectorization_row(state: ndarray, dim: int) -> ndarray:
    return ENGINE.reshape(state, (state.shape[0], dim, dim))


def _unvectorization_column(state: ndarray, dim: int) -> ndarray:
    # this should be equivalent and more broadly supported (e.g. by numba)
    # return ENGINE.reshape(state.T, (dim, dim, state.shape[0])).T
    return ENGINE.reshape(state, (state.shape[0], dim, dim), order="F")


def _unvectorization_system(state: ndarray, dim: int) -> ndarray:
    nqubits = int(np.log2(dim))
    axes_old = list(range(1, 2 * nqubits + 1))
    state = ENGINE.reshape(state, (state.shape[0],) + (2,) * 2 * nqubits)
    state = ENGINE.transpose(state, [0] + axes_old[1::2] + axes_old[0::2])
    return ENGINE.reshape(state, (state.shape[0],) + (2**nqubits,) * 2)


def _reshuffling(super_op: ndarray, ax1: int, ax2: int) -> ndarray:
    dim = int(np.sqrt(super_op.shape[0]))
    super_op = ENGINE.reshape(super_op, (dim,) * 4)
    axes = list(range(len(super_op.shape)))
    tmp = axes[ax1]
    axes[ax1] = axes[ax2]
    axes[ax2] = tmp
    super_op = ENGINE.transpose(super_op, axes)
    return ENGINE.reshape(super_op, [dim**2, dim**2])


def _random_statevector_real(dims: int):
    state = ENGINE.random.standard_normal(dims)
    return state / ENGINE.linalg.norm(state)


def _random_statevector(dims: int):
    state = ENGINE.random.standard_normal(dims)
    state = state + 1.0j * ENGINE.random.standard_normal(dims)
    return state / ENGINE.linalg.norm(state)


def _random_density_matrix_pure(dims: int):
    state = _random_statevector(dims)
    return ENGINE.outer(state, ENGINE.conj(state).T)


def _random_gaussian_matrix(dims: int, rank: int, mean: float, stddev: float):
    dims = (dims, rank)
    matrix = 1.0j * ENGINE.random.normal(loc=mean, scale=stddev, size=dims)
    matrix += ENGINE.random.normal(loc=mean, scale=stddev, size=dims)
    return matrix


def _random_density_matrix_hs_ginibre(dims: int, rank: int, mean: float, stddev: float):
    state = _random_gaussian_matrix(dims, rank, mean, stddev)
    state = ENGINE.matmul(state, ENGINE.transpose(ENGINE.conj(state), (1, 0)))
    return state / ENGINE.trace(state)


def _random_hermitian(dims: int):
    matrix = _random_gaussian_matrix(dims, dims, 0.0, 1.0)
    return (matrix + ENGINE.conj(matrix).T) / 2


def _random_hermitian_semidefinite(dims: int):
    matrix = _random_gaussian_matrix(dims, dims, 0.0, 1.0)
    return ENGINE.matmul(ENGINE.conj(matrix).T, matrix)


def _random_unitary(dims: int):
    H = _random_hermitian(dims)
    return expm(-1.0j * H / 2)


def _random_unitary_haar(dims: int):
    matrix = _random_gaussian_matrix(dims, dims, 0.0, 1.0)
    Q, R = ENGINE.linalg.qr(matrix)
    D = ENGINE.diag(R)
    D = D / ENGINE.abs(D)
    R = ENGINE.diag(D)
    return ENGINE.matmul(Q, R)


def _random_density_matrix_bures(dims: int, rank: int, mean: float, stddev: float):
    nqubits = int(np.log2(dims))
    state = ENGINE.eye(dims, dtype=ENGINE.complex128)
    state += _random_unitary(dims)
    state = ENGINE.matmul(
        state,
        _random_gaussian_matrix(dims, rank, mean, stddev),
    )
    state = ENGINE.matmul(state, ENGINE.transpose(ENGINE.conj(state), (1, 0)))
    return state / ENGINE.trace(state)


def _fill_tril(mat, symmetric):
    """Add symmetric random ints to off-diagonals"""
    dim = mat.shape[0]
    # Optimized for low dimensions
    if dim == 1:
        return

    if dim <= 4:
        mat[1, 0] = ENGINE.random.randint(2, dtype=ENGINE.uint8)
        if symmetric:
            mat[0, 1] = mat[1, 0]
        if dim > 2:
            mat[2, 0] = ENGINE.random.randint(2, dtype=np.uint8)
            mat[2, 1] = ENGINE.random.randint(2, dtype=np.uint8)
            if symmetric:
                mat[0, 2] = mat[2, 0]
                mat[1, 2] = mat[2, 1]
        if dim > 3:
            mat[3, 0] = ENGINE.random.randint(2, dtype=np.uint8)
            mat[3, 1] = ENGINE.random.randint(2, dtype=np.uint8)
            mat[3, 2] = ENGINE.random.randint(2, dtype=np.uint8)
            if symmetric:
                mat[0, 3] = mat[3, 0]
                mat[1, 3] = mat[3, 1]
                mat[2, 3] = mat[3, 2]
        return

    # Use numpy indices for larger dimensions
    rows, cols = ENGINE.tril_indices(dim, -1)
    vals = ENGINE.random.randint(2, size=rows.size, dtype=ENGINE.uint8)
    mat[(rows, cols)] = vals
    if symmetric:
        mat[(cols, rows)] = vals


def _inverse_tril(mat, block_inverse_threshold):
    """Invert a lower-triangular matrix with unit diagonal."""
    # Optimized inversion function for low dimensions
    dim = mat.shape[0]

    if dim <= 2:
        return mat

    if dim <= 5:
        inv = ENGINE.copy(mat)
        inv[2, 0] = mat[2, 0] ^ (mat[1, 0] & mat[2, 1])
        if dim > 3:
            inv[3, 1] = mat[3, 1] ^ (mat[2, 1] & mat[3, 2])
            inv[3, 0] = mat[3, 0] ^ (mat[3, 2] & mat[2, 0]) ^ (mat[1, 0] & inv[3, 1])
        if dim > 4:
            inv[4, 2] = (mat[4, 2] ^ (mat[3, 2] & mat[4, 3])) & 1
            inv[4, 1] = mat[4, 1] ^ (mat[4, 3] & mat[3, 1]) ^ (mat[2, 1] & inv[4, 2])
            inv[4, 0] = (
                mat[4, 0]
                ^ (mat[1, 0] & inv[4, 1])
                ^ (mat[2, 0] & inv[4, 2])
                ^ (mat[3, 0] & mat[4, 3])
            )
        return inv % 2

    # For higher dimensions we use Numpy's inverse function
    # however this function tends to fail and result in a non-symplectic
    # final matrix if n is too large.
    if dim <= block_inverse_threshold:
        return ENGINE.linalg.inv(mat) % 2

    # For very large matrices  we divide the matrix into 4 blocks of
    # roughly equal size and use the analytic formula for the inverse
    # of a block lower-triangular matrix:
    # inv([[A, 0],[C, D]]) = [[inv(A), 0], [inv(D).C.inv(A), inv(D)]]
    # call the inverse function recursively to compute inv(A) and invD

    dim1 = dim // 2
    mat_a = _inverse_tril(mat[0:dim1, 0:dim1], block_inverse_threshold)
    mat_d = _inverse_tril(mat[dim1:dim, dim1:dim], block_inverse_threshold)
    mat_c = (mat_d @ mat[dim1:dim, 0:dim1]) @ mat_a
    inv = ENGINE.vstack(
        [
            ENGINE.hstack([mat_a, ENGINE.zeros((dim1, dim - dim1), dtype=int)]),
            ENGINE.hstack([mat_c, mat_d]),
        ]
    )
    return inv % 2


def _sample_from_quantum_mallows_distribution(nqubits: int) -> tuple[ndarray, ndarray]:
    """Using the quantum Mallows distribution, samples a binary array
    representing a layer of Hadamard gates as well as an array with permutated
    qubit indexes. For more details, see Reference [1].

    Args:
        nqubits (int): number of qubits.

    Returns:
        (``ndarray``, ``ndarray`): tuple of binary ``ndarray`` and ``ndarray`` of indexes.

    Reference:
        1. S. Bravyi and D. Maslov, *Hadamard-free circuits expose the
            structure of the Clifford group*.
            `arXiv:2003.09412 [quant-ph] <https://arxiv.org/abs/2003.09412>`_.

    """
    exponents = ENGINE.arange(nqubits, 0, -1, dtype=ENGINE.int64)
    powers = 4**exponents
    powers[powers == 0] = ENGINE.iinfo(ENGINE.int64).max
    r = ENGINE.random.uniform(0, 1, size=nqubits)
    indexes = (-1) * ENGINE.ceil(np.log2(r + (1 - r) / powers)).astype(ENGINE.int64)
    idx_le_exp = indexes < exponents
    hadamards = idx_le_exp.astype(ENGINE.int64)
    idx_gt_exp = idx_le_exp ^ True
    indexes[idx_gt_exp] = 2 * exponents[idx_gt_exp] - indexes[idx_gt_exp] - 1
    permutations = np.empty(nqubits, dtype=int)
    mask = np.ones(nqubits, dtype=bool)
    for l, index in enumerate(indexes):
        available = np.flatnonzero(mask)
        permutations[l] = available[index]
        mask[permutations[l]] = False
    return hadamards, permutations


def _super_op_from_bcsz_measure_preamble(
    dims: int, rank: int
) -> tuple[ndarray, ndarray]:
    super_op = _random_gaussian_matrix(
        dims**2,
        rank=rank,
        mean=0.0,
        stddev=1.0,
    )
    super_op = super_op @ ENGINE.conj(super_op).T
    # partial trace implemented with einsum
    super_op_reduced = ENGINE.einsum("ijik->jk", ENGINE.reshape(super_op, (dims,) * 4))
    eigenvalues, eigenvectors = ENGINE.linalg.eigh(super_op_reduced)
    eigenvalues = ENGINE.sqrt(1.0 / eigenvalues)
    eigenvectors = eigenvectors.T
    operator = ENGINE.einsum("ij,ik->ijk", eigenvectors, ENGINE.conj(eigenvectors))
    operator = ENGINE.sum(
        eigenvalues.reshape(len(eigenvalues), 1, 1) * operator, axis=0
    )
    return operator, super_op


def _super_op_from_bcsz_measure_row(dims: int, rank: int) -> ndarray:
    operator, super_op = _super_op_from_bcsz_measure_preamble(dims, rank)
    operator = ENGINE.kron(ENGINE.eye(dims, dtype=operator.dtype), operator)
    return operator @ super_op @ operator


def _super_op_from_bcsz_measure_column(dims: int, rank: int) -> ndarray:
    operator, super_op = _super_op_from_bcsz_measure_preamble(dims, rank)
    operator = ENGINE.kron(operator, ENGINE.eye(dims, dtype=operator.dtype))
    return operator @ super_op @ operator


def _kraus_to_stinespring(
    kraus_ops: ndarray, initial_state_env: ndarray, dim_env: int
) -> ndarray:
    alphas = ENGINE.zeros((dim_env, dim_env, dim_env), dtype=kraus_ops.dtype)
    alphas[range(dim_env), range(dim_env)] = initial_state_env
    # batched kron product
    return ENGINE.einsum("aij,akl->ikjl", kraus_ops, alphas).reshape(
        2 * (kraus_ops.shape[1] * alphas.shape[1],)
    )


def _stinespring_to_kraus(
    stinespring: ndarray, initial_state_env: ndarray, dim: int, dim_env: int
) -> ndarray:
    stinespring = ENGINE.reshape(stinespring, (dim, dim_env, dim, dim_env))
    stinespring = ENGINE.swapaxes(stinespring, 1, 2)
    alphas = ENGINE.eye(dim_env, dtype=stinespring.dtype)
    stinespring = (alphas @ stinespring).reshape(dim, dim_env, dim + dim_env)
    stinespring = ENGINE.vstack(
        (stinespring[:, :, :dim_env], stinespring[:, :, dim_env:])
    )
    return (stinespring @ initial_state_env).reshape(dim, dim_env, dim_env)



import numpy as np

from qibo import gates, matrices
from qibo.config import raise_error
from qibo.quantum_info.linalg_operations import schmidt_decomposition

magic_basis = np.array(
    [[1, -1j, 0, 0], [0, 0, 1, -1j], [0, 0, -1, -1j], [1, 1j, 0, 0]]
) / np.sqrt(2)

bell_basis = np.array(
    [[1, 1, 0, 0], [0, 0, 1, 1], [0, 0, 1, -1], [1, -1, 0, 0]]
) / np.sqrt(2)

H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)


def u3_decomposition(unitary, backend):
    """Decomposes arbitrary one-qubit gates to U3.

    Args:
        unitary (ndarray): Unitary :math:`2 \\times 2` matrix to be decomposed.

    Returns:
        (float, float, float): parameters of U3 gate.
    """
    unitary = backend.cast(unitary)
    # https://github.com/Qiskit/qiskit-terra/blob/d2e3340adb79719f9154b665e8f6d8dc26b3e0aa/qiskit/quantum_info/synthesis/one_qubit_decompose.py#L221
    su2 = unitary / backend.np.sqrt(backend.np.linalg.det(unitary))
    theta = 2 * backend.np.arctan2(backend.np.abs(su2[1, 0]), backend.np.abs(su2[0, 0]))
    plus = backend.np.angle(su2[1, 1])
    minus = backend.np.angle(su2[1, 0])
    phi = plus + minus
    lam = plus - minus
    # explicit conversion to float to avoid issue on GPU
    return float(theta), float(phi), float(lam)


def calculate_psi(unitary, backend, magic_basis=magic_basis):
    """Solves the eigenvalue problem of :math:`U^{T} U`.

    See step (1) of Appendix A in arXiv:quant-ph/0011050.

    Args:
        unitary (ndarray): Unitary matrix of the gate we are decomposing
            in the computational basis.
        magic_basis (ndarray, optional): basis in which to solve the eigenvalue problem.
            Defaults to ``magic basis``.
        backend (:class:`qibo.backends.abstract.Backend`): Backend to use for calculations.

    Returns:
        ndarray: Eigenvectors in the computational basis and eigenvalues of :math:`U^{T} U`.
    """
    magic_basis = backend.cast(magic_basis)
    unitary = backend.cast(unitary)
    # write unitary in magic basis
    u_magic = (
        backend.np.transpose(backend.np.conj(magic_basis), (1, 0))
        @ unitary
        @ magic_basis
    )
    # construct and diagonalize UT_U
    ut_u = backend.np.transpose(u_magic, (1, 0)) @ u_magic
    ut_u_real = backend.np.real(ut_u) + backend.np.imag(ut_u)
    if backend.__class__.__name__ not in ("PyTorchBackend", "TensorflowBackend"):
        ut_u_real = np.round(ut_u_real, decimals=15)

    eigvals_real, psi_magic = backend.calculate_eigenvectors(ut_u_real, hermitian=True)
    # compute full eigvals as <psi|ut_u|psi>, as eigvals_real is only real
    eigvals = backend.np.sum(backend.np.conj(psi_magic) * (ut_u @ psi_magic), 0)
    # orthogonalize eigenvectors in the case of degeneracy (Gram-Schmidt)
    psi_magic, _ = backend.np.linalg.qr(psi_magic)
    # write psi in computational basis
    psi = backend.np.matmul(magic_basis, psi_magic)
    return psi, eigvals


def calculate_single_qubit_unitaries(psi, backend=None):
    """Calculates local unitaries that maps a maximally entangled basis to the magic basis.

    See Lemma 1 of Appendix A in arXiv:quant-ph/0011050.

    Args:
        psi (ndarray): Maximally entangled two-qubit states that define a basis.

    Returns:
        (ndarray, ndarray): Local unitaries UA and UB that map the given basis to the magic basis.
    """
    psi_magic = backend.np.matmul(backend.np.conj(backend.cast(magic_basis)).T, psi)
    if (
        backend.np.real(backend.calculate_matrix_norm(backend.np.imag(psi_magic)))
        > 1e-6
    ):  # pragma: no cover
        raise_error(NotImplementedError, "Given state is not real in the magic basis.")
    psi_bar = backend.cast(psi.T, copy=True)

    # find e and f by inverting (A3), (A4)
    ef = (psi_bar[0] + 1j * psi_bar[1]) / np.sqrt(2)
    e_f_ = (psi_bar[0] - 1j * psi_bar[1]) / np.sqrt(2)
    e, _, f = schmidt_decomposition(ef, [0], backend=backend)
    e, f = e[:, 0], f[0]
    e_, _, f_ = schmidt_decomposition(e_f_, [0], backend=backend)
    e_, f_ = e_[:, 0], f_[0]
    # find exp(1j * delta) using (A5a)
    ef_ = backend.np.kron(e, f_)
    phase = (
        1j
        * np.sqrt(2)
        * backend.np.sum(backend.np.multiply(backend.np.conj(ef_), psi_bar[2]))
    )
    v0 = backend.cast(np.asarray([1, 0]))
    v1 = backend.cast(np.asarray([0, 1]))
    # construct unitaries UA, UB using (A6a), (A6b)
    ua = backend.np.tensordot(v0, backend.np.conj(e), 0) + phase * backend.np.tensordot(
        v1, backend.np.conj(e_), 0
    )
    ub = backend.np.tensordot(v0, backend.np.conj(f), 0) + backend.np.conj(
        phase
    ) * backend.np.tensordot(v1, backend.np.conj(f_), 0)
    return ua, ub


def calculate_diagonal(unitary, ua, ub, va, vb, backend):
    """Calculates Ud matrix that can be written as exp(-iH).

    See Eq. (A1) in arXiv:quant-ph/0011050.
    Ud is diagonal in the magic and Bell basis.
    Also returns local unitaries that modify Ud so: pi/4 >= hx >= hy >= |hz|
    """
    # normalize U_A, U_B, V_A, V_B so that detU_d = 1
    # this is required so that sum(lambdas) = 0
    # and Ud can be written as exp(-iH)
    if backend.__class__.__name__ == "TensorflowBackend":  # pragma: no cover
        det = np.linalg.det(unitary) ** (1 / 16)
    else:
        det = backend.np.linalg.det(unitary) ** (1 / 16)
    ua *= det
    ub *= det
    va *= det
    vb *= det
    dag = lambda u: backend.np.transpose(backend.np.conj(u), (1, 0))
    u_dagger = dag(
        backend.np.kron(
            ua,
            ub,
        )
    )
    v_dagger = dag(backend.np.kron(va, vb))
    ud = u_dagger @ unitary @ v_dagger

    lambdas = to_bell_diagonal(ud, backend)

    # We permute to ensure we will have hx >= hy >= hz in the end
    hx, hy, hz = calculate_h_vector(lambdas, backend)

    # 1. force coefficients to be in pi/4 >= ... >= 0 interval, using pi/2 periodicity and pi/4 symmetry
    fit = lambda alpha: min(alpha % (np.pi / 2), np.pi / 2 - (alpha % (np.pi / 2)))

    # 2. permute to ensure ordering:
    alphas_ordered = sorted(
        [[hx, 0], [hy, 1], [hz, 2]], key=lambda x: fit(x[0]), reverse=True
    )

    correction = {
        "left_A": matrices.I.copy(),
        "left_B": matrices.I.copy(),
        "right_A": matrices.I.copy(),
        "right_B": matrices.I.copy(),
    }

    permutation = [x[1] for x in alphas_ordered]

    if permutation[0] == 1:
        correction["left_A"] @= matrices.S
        correction["left_B"] @= matrices.S
        correction["right_A"] = dag(matrices.S) @ correction["right_A"]
        correction["right_B"] = dag(matrices.S) @ correction["right_B"]

    elif permutation[0] == 2:
        correction["left_A"] = correction["left_A"] @ matrices.H
        correction["left_B"] = correction["left_B"] @ matrices.H
        correction["right_A"] = dag(matrices.H) @ correction["right_A"]
        correction["right_B"] = dag(matrices.H) @ correction["right_B"]

    if not (permutation[1] == 1 or permutation[2] == 2):
        correction["left_A"] = correction["left_A"] @ (
            matrices.S @ matrices.H @ matrices.S
        )
        correction["left_B"] = correction["left_B"] @ (
            matrices.S @ matrices.H @ matrices.S
        )
        correction["right_A"] = (
            dag(matrices.S @ matrices.H @ matrices.S) @ correction["right_A"]
        )
        correction["right_B"] = (
            dag(matrices.S @ matrices.H @ matrices.S) @ correction["right_B"]
        )

    # 3. find local corrections to enforce, as possible, conditions on h
    paulis = ["X", "Y", "Z"]
    for i, (alpha, _) in enumerate(alphas_ordered):
        if i < 2:
            if alpha < 0:
                alpha += np.pi / 2
                correction["left_A"] = correction["left_A"] @ (
                    1j * getattr(matrices, paulis[i])
                )
                correction["left_B"] = correction["left_B"] @ getattr(
                    matrices, paulis[i]
                )
            if (
                alpha > np.pi / 4
            ):  # can't conjugate so sign of z alternates to compensate
                # Swap alpha_i and alpha_z sign
                correction["left_B"] = correction["left_B"] @ getattr(
                    matrices, paulis[(i + 1) % 2]
                )
                correction["right_B"] = (
                    getattr(matrices, paulis[(i + 1) % 2]) @ correction["right_B"]
                )
                # Add pi/2 to alpha_i
                correction["left_A"] = correction["left_A"] @ (
                    1j * getattr(matrices, paulis[i])
                )
                correction["left_B"] = correction["left_B"] @ getattr(
                    matrices, paulis[i]
                )
        elif abs(alpha) > np.pi / 4:
            correction["left_A"] = correction["left_A"] @ (
                (1j if alpha < 0 else -1j) * getattr(matrices, paulis[i])
            )
            correction["left_B"] = correction["left_B"] @ getattr(matrices, paulis[i])

    # 4. apply corrections
    correction["left_A"] = backend.cast(correction["left_A"])
    correction["left_B"] = backend.cast(correction["left_B"])
    correction["right_A"] = backend.cast(correction["right_A"])
    correction["right_B"] = backend.cast(correction["right_B"])
    ua = backend.np.matmul(ua, correction["left_A"])
    ub = backend.np.matmul(ub, correction["left_B"])
    va = backend.np.matmul(correction["right_A"], va)
    vb = backend.np.matmul(correction["right_B"], vb)
    ud = backend.np.matmul(
        backend.np.kron(dag(correction["left_A"]), dag(correction["left_B"])),
        backend.np.matmul(
            ud, backend.np.kron(dag(correction["right_A"]), dag(correction["right_B"]))
        ),
    )
    return ua, ub, ud, va, vb


def magic_decomposition(unitary, backend=None):
    """Decomposes an arbitrary unitary to (A1) from arXiv:quant-ph/0011050."""

    unitary = backend.cast(unitary)
    psi, eigvals = calculate_psi(unitary, backend=backend)
    psi_tilde = backend.np.conj(backend.np.sqrt(eigvals)) * backend.np.matmul(
        unitary, psi
    )
    va, vb = calculate_single_qubit_unitaries(psi, backend=backend)
    ua_dagger, ub_dagger = calculate_single_qubit_unitaries(psi_tilde, backend=backend)
    dag = lambda U: backend.np.transpose(backend.np.conj(U), (1, 0))
    ua, ub = dag(ua_dagger), dag(ub_dagger)
    return calculate_diagonal(unitary, ua, ub, va, vb, backend=backend)


def to_bell_diagonal(ud, backend, bell_basis=bell_basis):
    """Transforms a matrix to the Bell basis and checks if it is diagonal."""
    ud = backend.cast(ud)
    bell_basis = backend.cast(bell_basis)

    ud_bell = (
        backend.np.transpose(backend.np.conj(bell_basis), (1, 0)) @ ud @ bell_basis
    )
    ud_diag = backend.np.diag(ud_bell)
    if not backend.np.allclose(
        backend.np.diag(ud_diag), ud_bell, atol=1e-6, rtol=1e-6
    ):  # pragma: no cover
        return None
    uprod = backend.to_numpy(backend.np.prod(ud_diag))
    if not np.allclose(uprod, 1.0, atol=1e-6, rtol=1e-6):  # pragma: no cover
        return None
    return ud_diag


def calculate_h_vector(ud_diag, backend):
    """Finds h parameters corresponding to exp(-iH).

    See Eq. (4)-(5) in arXiv:quant-ph/0307177.
    """
    lambdas = -backend.np.angle(ud_diag)
    hx = (lambdas[0] + lambdas[2]) / 2.0
    hy = (lambdas[1] + lambdas[2]) / 2.0
    hz = (lambdas[0] + lambdas[1]) / 2.0
    return hx, hy, hz


def cnot_decomposition(q0, q1, hx, hy, hz, backend):
    """Performs decomposition (6) from arXiv:quant-ph/0307177."""
    h = backend.cast(H)
    u3 = backend.cast(-1j * matrices.H)
    # use corrected version from PRA paper (not arXiv)
    u2 = -u3 @ gates.RX(0, 2 * hx - np.pi / 2).matrix(backend)
    # add an extra exp(-i pi / 4) global phase to get exact match
    v2 = np.exp(-1j * np.pi / 4) * gates.RZ(0, 2 * hz).matrix(backend)
    v3 = gates.RZ(0, -2 * hy).matrix(backend)
    w = backend.cast((matrices.I - 1j * matrices.X) / np.sqrt(2))
    # change CNOT to CZ using Hadamard gates
    return [
        gates.H(q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(h @ v2 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u3, q0),
        gates.Unitary(h @ v3 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(backend.np.conj(w).T @ h, q1),
    ]


def cnot_decomposition_light(q0, q1, hx, hy, backend):
    """Performs decomposition (24) from arXiv:quant-ph/0307177."""
    h = backend.cast(H)
    w = backend.cast((matrices.I - 1j * matrices.X) / np.sqrt(2))
    u2 = gates.RX(0, 2 * hx).matrix(backend)
    v2 = gates.RZ(0, -2 * hy).matrix(backend)
    # change CNOT to CZ using Hadamard gates
    return [
        gates.Unitary(backend.np.conj(w).T, q0),
        gates.Unitary(h @ w, q1),
        gates.CZ(q0, q1),
        gates.Unitary(u2, q0),
        gates.Unitary(h @ v2 @ h, q1),
        gates.CZ(q0, q1),
        gates.Unitary(w, q0),
        gates.Unitary(backend.np.conj(w).T @ h, q1),
    ]


def _get_z_component(unitary, backend):
    """Calculates the hz component from a unitary's magic decomposition."""
    _, _, ud, _, _ = magic_decomposition(unitary, backend=backend)
    ud_diag = to_bell_diagonal(ud, backend=backend)
    _, _, hz = calculate_h_vector(ud_diag, backend=backend)
    return float(hz)


def _two_qubit_decomposition_without_z(q0, q1, unitary, backend):
    """Implements Theorem 2 decomposition (2 CNOTs) for hz=0 case."""
    # Get magic decomposition
    u4, v4, ud, u1, v1 = magic_decomposition(unitary, backend=backend)
    ud_diag = to_bell_diagonal(ud, backend=backend)
    hx, hy, _ = calculate_h_vector(ud_diag, backend=backend)
    hx, hy = float(hx), float(hy)

    # Get light decomposition
    gatelist = cnot_decomposition_light(q0, q1, hx, hy, backend=backend)
    # Combine with initial and final local unitaries
    g0, g1 = gatelist[:2]
    gatelist[0] = gates.Unitary(backend.cast(g0.parameters[0]) @ u1, q0)
    gatelist[1] = gates.Unitary(backend.cast(g1.parameters[0]) @ v1, q1)

    g0, g1 = gatelist[-2:]
    gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
    gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    return gatelist


def _two_qubit_decomposition_with_z(q0, q1, unitary, backend):
    """Implements Theorem 1 decomposition (3 CNOTs) for hz≠0 case."""
    # Get magic decomposition
    u4, v4, ud, u1, v1 = magic_decomposition(unitary, backend=backend)
    ud_diag = to_bell_diagonal(ud, backend=backend)
    hx, hy, hz = calculate_h_vector(ud_diag, backend=backend)
    hx, hy, hz = float(hx), float(hy), float(hz)

    # Get full decomposition
    cnot_dec = cnot_decomposition(q0, q1, hx, hy, hz, backend=backend)

    # Combine with initial and final local unitaries
    gatelist = [
        gates.Unitary(u1, q0),
        gates.Unitary(backend.cast(H) @ v1, q1),
    ]
    gatelist.extend(cnot_dec[1:])
    g0, g1 = gatelist[-2:]
    gatelist[-2] = gates.Unitary(u4 @ g0.parameters[0], q0)
    gatelist[-1] = gates.Unitary(v4 @ g1.parameters[0], q1)

    return gatelist


def two_qubit_decomposition(q0, q1, unitary, backend, threshold=1e-6):
    """Performs two qubit unitary gate decomposition.

    Args:
        q0 (int): index of the first qubit.
        q1 (int): index of the second qubit.
        unitary (ndarray): Unitary :math:`4 \\times 4` to be decomposed.
        backend (:class:`qibo.backends.Backend`): Backend to use for calculations.
        threshold (float): Threshold for determining if hz component is zero.

    Returns:
        list: gates implementing the decomposition
    """
    # Handle identity case efficiently
    if backend.np.allclose(
        unitary, backend.identity_density_matrix(nqubits=2, normalize=False)
    ):
        return []

    z_component = _get_z_component(unitary, backend)
    if abs(z_component) < threshold:
        return _two_qubit_decomposition_without_z(q0, q1, unitary, backend)
    return _two_qubit_decomposition_with_z(q0, q1, unitary, backend)



"""Module defining the Hamming-weight-preserving backend."""

import inspect
from functools import cache
from importlib.util import find_spec, module_from_spec

from qibo.config import raise_error


def HammingWeightBackend(engine=None):
    """Dynamically create a HammingWeightBackend class based on the selected backend."""

    from qibo.backends import construct_backend  # pylint: disable=C415

    if engine is None:
        from qibo.backends import (  # pylint: disable=C0415
            _check_backend,
            _get_engine_name,
        )

        engine = _get_engine_name(_check_backend(engine))

    backend = None  # needed for pylint
    if engine == "numpy":
        backend = construct_backend("numpy", platform=engine)
    elif engine in ["numba", "cupy", "cuquantum"]:
        backend = construct_backend("qibojit", platform=engine)
    elif engine in ["tensorflow", "pytorch"]:  # pragma: no cover
        backend = construct_backend("qiboml", platform=engine)
    else:  # pragma: no cover
        raise_error(
            NotImplementedError,
            f"Backend `{engine}` is not supported for "
            + "Hamming-weight-preserving circuit simulation.",
        )

    import qibo.backends._hamming_weight_operations as module

    methods = {
        name: func for name, func in inspect.getmembers(module, inspect.isfunction)
    }

    for method_name in ["_get_cached_strings", "_get_lexicographical_order"]:
        if method_name in methods:
            methods[method_name] = cache(methods[method_name])

    HWBackend = type(
        "HammingWeightBackend", (backend.__class__,), methods
    )  # pylint: disable=E0606

    hw_backend = HWBackend()
    hw_backend.name = "hamming_weight"
    hw_backend.platform = engine

    hw_backend.calculate_full_probabilities = backend.calculate_probabilities

    hw_backend._dict_cached_strings_one = {}
    hw_backend._dict_cached_strings_two = {}

    hw_backend._dict_indexes = None

    return hw_backend



from enum import EnumMeta, Flag, auto
from functools import reduce
from operator import or_

from qibo import gates
from qibo.backends import _check_backend
from qibo.config import log, raise_error
from qibo.models import Circuit
from qibo.transpiler._exceptions import DecompositionError
from qibo.transpiler.decompositions import (
    cnot_dec_temp,
    cz_dec,
    gpi2_dec,
    iswap_dec,
    opt_dec,
    u3_dec,
)


class FlagMeta(EnumMeta):
    """Metaclass for :class:`qibo.transpiler.unroller.NativeGates` that allows initialization with a list of gate name strings."""

    def __getitem__(cls, keys):
        if isinstance(keys, str):
            try:
                return super().__getitem__(keys)
            except KeyError:
                return super().__getitem__("NONE")
        return reduce(or_, [cls[key] for key in keys])  # pylint: disable=E1136


class NativeGates(Flag, metaclass=FlagMeta):
    """Define native gates supported by the unroller. A native gate set should contain at least
    one two-qubit gate (:class:`qibo.gates.gates.CZ` or :class:`qibo.gates.gates.iSWAP`),
    and at least one single-qubit gate
    (:class:`qibo.gates.gates.GPI2` or :class:`qibo.gates.gates.U3`).

    Possible gates are:
        - :class:`qibo.gates.gates.I`
        - :class:`qibo.gates.gates.Z`
        - :class:`qibo.gates.gates.RZ`
        - :class:`qibo.gates.gates.M`
        - :class:`qibo.gates.gates.GPI2`
        - :class:`qibo.gates.gates.U3`
        - :class:`qibo.gates.gates.CZ`
        - :class:`qibo.gates.gates.iSWAP`
        - :class:`qibo.gates.gates.CNOT`
    """

    NONE = 0
    I = auto()
    Z = auto()
    RZ = auto()
    M = auto()
    GPI2 = auto()
    U3 = auto()
    CZ = auto()
    iSWAP = auto()
    CNOT = auto()  # For testing purposes

    @classmethod
    def default(cls):
        """Return default native gates set."""
        return cls.CZ | cls.GPI2 | cls.I | cls.Z | cls.RZ | cls.M

    @classmethod
    def from_gatelist(cls, gatelist: list):
        """Create a NativeGates object containing all gates from a ``gatelist``."""
        natives = cls(0)
        for gate in gatelist:
            natives |= cls.from_gate(gate)
        return natives

    @classmethod
    def from_gate(cls, gate):
        """Create a :class:`qibo.transpiler.unroller.NativeGates`
        object from a :class:`qibo.gates.gates.Gate`."""
        if isinstance(gate, gates.Gate):
            return cls.from_gate(gate.__class__)
        try:
            return getattr(cls, gate.__name__)
        except AttributeError:
            raise_error(ValueError, f"Gate {gate} cannot be used as native.")


# TODO: Make setting single-qubit native gates more flexible
class Unroller:
    """Decomposes a circuit to native gates."""

    def __init__(
        self,
        native_gates: NativeGates,
        backend=None,
    ):
        self.native_gates = native_gates
        self.backend = backend
        """Initializes the unroller.

        Args:
            native_gates (:class:`qibo.transpiler.unroller.NativeGates`): Native gates to use in the transpiled circuit.
            backend (:class:`qibo.backends.Backend`): Backend to use for gate matrix.
        """

    def __call__(self, circuit: Circuit):
        """Decomposes a circuit to native gates.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be decomposed.

        Returns:
            (:class:`qibo.models.circuit.Circuit`): Decomposed circuit.
        """
        translated_circuit = Circuit(**circuit.init_kwargs)
        for gate in circuit.queue:
            translated_circuit.add(
                translate_gate(
                    gate,
                    self.native_gates,
                    backend=self.backend,
                )
            )
        return translated_circuit


def translate_gate(
    gate,
    native_gates: NativeGates,
    backend=None,
):
    """Maps gates to a hardware-native implementation.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Gate to be decomposed.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`): Native gates supported by the hardware.
        backend (:class:`qibo.backends.Backend`): Backend to use for gate matrix.

    Returns:
        list: Native gates that decompose the input gate.
    """
    backend = _check_backend(backend)

    if isinstance(gate, (gates.I, gates.Align)):
        return gate

    if isinstance(gate, gates.M):
        gate.basis_gates = len(gate.basis_gates) * [gates.Z]
        gate.basis = []
        return gate

    if len(gate.qubits) == 1:
        return _translate_single_qubit_gates(gate, native_gates, backend)

    decomposition_2q = _translate_two_qubit_gates(gate, native_gates, backend)
    final_decomposition = []
    for decomposed_2q_gate in decomposition_2q:
        if len(decomposed_2q_gate.qubits) == 1:
            final_decomposition += _translate_single_qubit_gates(
                decomposed_2q_gate, native_gates, backend
            )
        else:
            final_decomposition.append(decomposed_2q_gate)
    return final_decomposition


def _translate_single_qubit_gates(
    gate: gates.Gate, single_qubit_natives: NativeGates, backend
):
    """Helper method for :meth:`translate_gate`.

    Maps single-qubit gates to a hardware-native implementation.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Gate to be decomposed.
        single_qubit_natives (:class:`qibo.transpiler.unroller.NativeGates`): Single qubit native gates supported by the hardware.
        backend (:class:`qibo.backends.Backend`): Backend to use for gate matrix.

    Returns:
        list: Native gates that decompose the input gate.
    """
    if NativeGates.U3 & single_qubit_natives:
        return u3_dec(gate, backend)

    if NativeGates.GPI2 & single_qubit_natives:
        return gpi2_dec(gate, backend)

    raise_error(DecompositionError, "Use U3 or GPI2 as single qubit native gates")


def _translate_two_qubit_gates(gate: gates.Gate, native_gates: NativeGates, backend):
    """Helper method for :meth:`translate_gate`.

    Maps two-qubit gates to a hardware-native implementation.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Gate to be decomposed.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`): Native gates
            supported by the hardware.
        backend (:class:`qibo.backends.Backend`): Backend to use for gate matrix.

    Returns:
        list: Native gates that decompose the input gate.
    """
    if (
        native_gates & (NativeGates.CZ | NativeGates.iSWAP)
    ) is NativeGates.CZ | NativeGates.iSWAP:
        # Check for a special optimized decomposition.
        if gate.__class__ in opt_dec.decompositions:
            return opt_dec(gate, backend)

        # Check if the gate has a CZ decomposition
        if gate.__class__ not in iswap_dec.decompositions:
            return cz_dec(gate, backend)

        # Check the decomposition with less 2 qubit gates.
        if cz_dec.count_2q(gate, backend) < iswap_dec.count_2q(gate, backend):
            return cz_dec(gate)

        if cz_dec.count_2q(gate, backend) > iswap_dec.count_2q(gate, backend):
            return iswap_dec(gate, backend)

        # If equal check the decomposition with less 1 qubit gates.
        # This is never used for now but may be useful for future generalization
        if cz_dec.count_1q(gate, backend) < iswap_dec.count_1q(
            gate, backend
        ):  # pragma: no cover
            return cz_dec(gate, backend)

        return iswap_dec(gate, backend)  # pragma: no cover

    if native_gates & NativeGates.CZ:
        return cz_dec(gate, backend)

    if native_gates & NativeGates.iSWAP:
        if gate.__class__ in iswap_dec.decompositions:
            return iswap_dec(gate, backend)

        # First decompose into CZ
        cz_decomposed = cz_dec(gate, backend)
        # Then CZ are decomposed into iSWAP
        iswap_decomposed = []
        for g in cz_decomposed:
            # Need recursive function as gates.Unitary is not in iswap_dec
            for g_translated in translate_gate(
                g, native_gates=native_gates, backend=backend
            ):
                iswap_decomposed.append(g_translated)
        return iswap_decomposed

    # For testing purposes
    # No CZ, iSWAP gates in the native gate set
    # Decompose CNOT, CZ, SWAP gates into CNOT gates
    if native_gates & NativeGates.CNOT:
        return cnot_dec_temp(gate, backend)

    raise_error(
        DecompositionError,
        "Use only CZ and/or iSWAP as native gates. CNOT is allowed in circuits where the two-qubit gates are limited to CZ, CNOT, and SWAP.",
    )  # pragma: no cover



from functools import reduce
from typing import Optional, Union

import numpy as np

from qibo import symbols
from qibo.backends import Backend, _check_backend
from qibo.config import raise_error
from qibo.hamiltonians.hamiltonians import Hamiltonian, SymbolicHamiltonian


def X(nqubits, dense: bool = True, backend=None):
    """Non-interacting Pauli-:math:`X` Hamiltonian.

    .. math::
        H = - \\sum _{k=0}^N \\, X_{k} \\, .

    Args:
        nqubits (int): number of qubits.
        dense (bool, optional): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
            Defaults to ``True``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """
    return _OneBodyPauli(nqubits, symbols.X, dense, backend=backend)


def Y(nqubits, dense: bool = True, backend=None):
    """Non-interacting Pauli-:math:`Y` Hamiltonian.

    .. math::
        H = - \\sum _{k=0}^{N} \\, Y_{k} \\, .

    Args:
        nqubits (int): number of qubits.
        dense (bool): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """
    return _OneBodyPauli(nqubits, symbols.Y, dense, backend=backend)


def Z(nqubits, dense: bool = True, backend=None):
    """Non-interacting Pauli-:math:`Z` Hamiltonian.

    .. math::
        H = - \\sum _{k=0}^{N} \\, Z_{k} \\, .

    Args:
        nqubits (int): number of qubits.
        dense (bool): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """
    return _OneBodyPauli(nqubits, symbols.Z, dense, backend=backend)


def TFIM(nqubits, h: float = 0.0, dense: bool = True, backend=None):
    """Transverse field Ising model with periodic boundary conditions.

    .. math::
        H = - \\sum _{k=0}^{N} \\, \\left(Z_{k} \\, Z_{k + 1} + h \\, X_{k}\\right) \\, .

    Args:
        nqubits (int): number of qubits.
        h (float, optional): value of the transverse field. Defaults to :math:`0.0`.
        dense (bool, optional): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
            Defaults to ``True``.
    """
    if nqubits < 2:
        raise_error(ValueError, "Number of qubits must be larger than one.")
    backend = _check_backend(backend)
    if dense:
        condition = lambda i, j: i in {j % nqubits, (j + 1) % nqubits}
        ham = -_build_spin_model(nqubits, backend.matrices.Z, condition, backend)
        if h != 0:
            condition = lambda i, j: i == j % nqubits
            ham -= h * _build_spin_model(
                nqubits, backend.matrices.X, condition, backend
            )
        return Hamiltonian(nqubits, ham, backend=backend)

    term = lambda q1, q2: symbols.Z(q1, backend=backend) * symbols.Z(
        q2, backend=backend
    ) + h * symbols.X(q1, backend=backend)
    form = -1 * sum(term(i, i + 1) for i in range(nqubits - 1)) - term(nqubits - 1, 0)
    ham = SymbolicHamiltonian(form=form, nqubits=nqubits, backend=backend)
    return ham


def MaxCut(
    nqubits,
    dense: bool = True,
    adj_matrix: Optional[Union[list[list[float]], np.ndarray]] = None,
    backend: Optional[Backend] = None,
):
    """Max Cut Hamiltonian.

    .. math::
        H = -\\frac{1}{2} \\, \\sum _{j, k = 0}^{N}  \\, \\left(1 - Z_{j} \\, Z_{k}\\right) \\, .

    Args:
        nqubits (int): number of qubits.
        dense (bool): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
        adj_matrix (list[list[float]] | np.ndarray): Adjecency matrix of the graph. Defaults to a
            homogeneous fully connected graph with all edges having an equal 1.0 weight.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.
    """
    if adj_matrix is None:
        adj_matrix = np.ones((nqubits, nqubits))
    elif len(adj_matrix) != nqubits:
        raise_error(
            RuntimeError,
            f"Expected an adjacency matrix of shape ({nqubits},{nqubits}) for a {nqubits}-qubits system.",
        )

    form = -sum(
        adj_matrix[i][j]
        * (1 - symbols.Z(i, backend=backend) * symbols.Z(j, backend=backend))
        for i in range(nqubits)
        for j in range(nqubits)
    )
    form /= 2

    ham = SymbolicHamiltonian(form, nqubits=nqubits, backend=backend)

    if dense:
        return ham.dense

    return ham


def LABS(nqubits: int, dense: bool = True, backend: Optional[Backend] = None):
    """Create Hamiltonian of the Low Autocorrelation Binary Sequences (LABS) problem.

    Given an integer :math:`n > 2`, the LABS problem consists of finding a binary sequence
    :math:`b \\in \\{0, \\, 1\\}^{n}` the minimizes the *sidelobe energy* :math:`E(b)`
    defined as

    .. math::
        E(b) = \\sum_{j=1}^{n-1} \\, C_{j}^{2}(b) \\, ;
        \\quad C_{j}(b) = \\sum_{k=0}^{n-j} \\, b_{k} \\, b_{k + j} \\, ,

    where :math:`C_{j}(b)` is the :math:`j`-th *autocorrelation* of :math:`b`.

    Args:
        nqubits (int): Total number of qubits.
        dense (bool): If ``True`` it creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    References:
        1. T. Packebusch and S. Mertens, *Low autocorrelation binary sequences*,
        `J. Phys. A: Math. Theor. 49 (2016) 165001
        <https://doi.org/10.1088/1751-8113/49/16/165001>`_.
    """
    if nqubits < 2:
        raise_error(
            ValueError,
            f"LABS problem only defined for ``nqubits > 2``, but ``nqubits = {nqubits}``.",
        )

    hamiltonian = 0
    for ind in range(1, nqubits):
        term = sum(
            symbols.Z(qubit, backend=backend) * symbols.Z(qubit + ind, backend=backend)
            for qubit in range(nqubits - ind)
        )
        hamiltonian += term**2

    hamiltonian = SymbolicHamiltonian(hamiltonian, nqubits=nqubits, backend=backend)

    if dense:
        return hamiltonian.dense

    return hamiltonian


def Heisenberg(
    nqubits,
    coupling_constants: Union[float, int, list, tuple],
    external_field_strengths: Union[float, int],
    dense: bool = True,
    backend=None,
):
    """Heisenberg model on a :math:`1`-dimensional periodic lattice.

    The general :math:`n`-qubit Hamiltonian is given by

    .. math::
        \\begin{align}
        H &= -\\sum_{k = 1}^{n} \\, \\left(
            J_{x} \\, X_{k} \\, X_{k + 1}
            + J_{y} \\, Y_{k} \\, Y_{k + 1}
            + J_{z} \\, Z_{k} \\, Z_{k + 1} \\right) \\\\
        &\\quad\\,\\, - \\sum_{k = 1}^{n} \\left(
            h_{x} \\, X_{k} + h_{y} \\, Y_{k} + h_{z} \\, Z_{k}
            \\right) \\, ,
        \\end{align}

    where :math:`\\{J_{x}, \\, J_{y}, \\, J_{z}\\}` are called the ``coupling constants``,
    :math:`\\{h_{x}, \\, h_{y}, \\, h_{z}\\}` are called the ``external field strengths``,
    and :math:`\\{X, \\, Y, \\, Z\\}` are the usual Pauli operators.

    Args:
        nqubits (int): number of qubits.
        coupling_constants (list or tuple or float or int): list or tuple with the
            three coupling constants :math:`\\{J_{x}, J_{y}, J{z}\\}`.
            If ``int`` or ``float``, then :math:`J_{x} = J_{y} = J_{z}`.
        external_field_strength (list or tuple or float or int): list or tuple with the
            external magnetic field strengths :math:`\\{h_{x}, \\, h_{y}, \\, h_{z}\\}`.
            If ``int`` or ``float``, then :math:`h_{x} = h_{y} = h_{z}`.
        dense (bool, optional): If ``True``, creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
            Defaults to ``True``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        :class:`qibo.hamiltonians.Hamiltonian` or :class:`qibo.hamiltonians.SymbolicHamiltonian`:
        Heisenberg Hamiltonian.
    """
    if isinstance(coupling_constants, (list, tuple)) and len(coupling_constants) != 3:
        raise_error(
            ValueError,
            f"When `coupling_constants` is type `int` or `list`, it must have length == 3.",
        )

    if isinstance(coupling_constants, (float, int)):
        coupling_constants = [coupling_constants] * 3

    if (
        isinstance(external_field_strengths, (list, tuple))
        and len(external_field_strengths) != 3
    ):
        raise_error(
            ValueError,
            f"When `external_field_strengths` is type `int` or `list`, it must have length == 3.",
        )

    if isinstance(external_field_strengths, (float, int)):
        external_field_strengths = [external_field_strengths] * 3

    backend = _check_backend(backend)

    paulis = (symbols.X, symbols.Y, symbols.Z)

    if dense:
        condition = lambda i, j: i in {j % nqubits, (j + 1) % nqubits}
        matrix = np.zeros((2**nqubits, 2**nqubits), dtype=complex)
        matrix = backend.cast(matrix, dtype=matrix.dtype)
        for ind, pauli in enumerate(paulis):
            double_term = _build_spin_model(
                nqubits, pauli(0, backend=backend).matrix, condition, backend
            )
            double_term = backend.cast(double_term, dtype=double_term.dtype)
            matrix = matrix - coupling_constants[ind] * double_term
            matrix = (
                matrix
                + external_field_strengths[ind]
                * _OneBodyPauli(nqubits, pauli, dense, backend).matrix
            )

        return Hamiltonian(nqubits, matrix, backend=backend)

    def h(symbol):
        return lambda q1, q2: symbol(q1, backend=backend) * symbol(q2, backend=backend)

    def term(q1, q2):
        return sum(
            coeff * h(operator)(q1, q2)
            for coeff, operator in zip(coupling_constants, paulis)
        )

    form = -1 * sum(term(i, i + 1) for i in range(nqubits - 1)) - term(nqubits - 1, 0)
    form -= sum(
        field_strength * pauli(qubit)
        for qubit in range(nqubits)
        for field_strength, pauli in zip(external_field_strengths, paulis)
        if field_strength != 0.0
    )

    ham = SymbolicHamiltonian(form=form, backend=backend)
    return ham


def XXX(
    nqubits,
    coupling_constant: Union[float, int] = 1,
    external_field_strengths: Union[float, int, list, tuple] = [0.5, 0, 0],
    dense: bool = True,
    backend=None,
):
    """Heisenberg :math:`\\mathrm{XXX}` model with periodic boundary conditions.

    The :math:`n`-qubit :math:`\\mathrm{XXX}` Hamiltonian is given by

    .. math::
        H = -J \\, \\sum_{k = 1}^{n} \\, \\left(
            X_{k} \\, X_{k + 1} + Y_{k} \\, Y_{k + 1} + Z_{k} \\, Z_{k + 1}
            \\right)
            - \\sum_{k = 1}^{n} \\left(h_{x} \\, X_{k} + h_{y} \\, Y_{k}
            + h_{z} \\, Z_{k} \\right) \\, ,

    where :math:`J` is the ``coupling_constant``, :math:`\\{h_{x}, \\, h_{y}, \\, h_{z}\\}`
    are called the ``external field strengths``, and :math:`\\{X, \\, Y, \\, Z\\}` are the usual
    Pauli operators.

    Args:
        nqubits (int): number of qubits.
        coupling_constant (float or int, optional): coupling constant :math:`J`.
            Defaults to :math:`1`.
        external_field_strengths (float or int or list or tuple, optional): list or tuple with the
            external magnetic field strengths :math:`\\{h_{x}, \\, h_{y}, \\, h_{z}\\}`.
            If ``int`` or ``float``, then :math:`h_{x} = h_{y} = h_{z}`.
            Defaults to :math:`[1/2, \\, 0, \\, 0]`.
        dense (bool, optional): If ``True``, creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
            Defaults to ``True``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        :class:`qibo.hamiltonians.Hamiltonian` or :class:`qibo.hamiltonians.SymbolicHamiltonian`:
        Heisenberg :math:`\\mathrm{XXX}` Hamiltonian.
    """
    if not isinstance(coupling_constant, (float, int)):
        raise_error(
            TypeError,
            "`coupling_constant` must be type float or int, "
            + f"but it is type {type(coupling_constant)}.",
        )

    return Heisenberg(
        nqubits,
        coupling_constants=coupling_constant,
        external_field_strengths=external_field_strengths,
        dense=dense,
        backend=backend,
    )


def XXZ(nqubits, delta=0.5, dense: bool = True, backend=None):
    """Heisenberg :math:`\\mathrm{XXZ}` model with periodic boundary conditions.

    .. math::
        H = \\sum _{k=0}^N \\, \\left( X_{k} \\, X_{k + 1} + Y_{k} \\, Y_{k + 1}
            + \\delta Z_{k} \\, Z_{k + 1} \\right) \\, .

    Example:
        .. testcode::

            from qibo.hamiltonians import XXZ
            h = XXZ(3) # initialized XXZ model with 3 qubits

    Args:
        nqubits (int): number of qubits.
        delta (float, optional): coefficient for the :math:`Z` component.
            Defaults to :math:`0.5`.
        dense (bool, optional): If ``True``, creates the Hamiltonian as a
            :class:`qibo.core.hamiltonians.Hamiltonian`, otherwise it creates
            a :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
            Defaults to ``True``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        :class:`qibo.hamiltonians.Hamiltonian` or :class:`qibo.hamiltonians.SymbolicHamiltonian`:
        Heisenberg :math:`\\mathrm{XXZ}` Hamiltonian.
    """
    if nqubits < 2:
        raise_error(ValueError, "Number of qubits must be larger than one.")

    return Heisenberg(nqubits, [-1, -1, -delta], 0, dense=dense, backend=backend)


def _multikron(matrix_list, backend):
    """Calculates Kronecker product of a list of matrices.

    Args:
        matrix_list (list): List of matrices as ``ndarray``.

    Returns:
        ndarray: Kronecker product of all matrices in ``matrix_list``.
    """
    return reduce(backend.np.kron, matrix_list)


def _build_spin_model(nqubits, matrix, condition, backend):
    """Helper method for building nearest-neighbor spin model Hamiltonians."""
    h = sum(
        reduce(
            backend.np.kron,
            (
                matrix if condition(i, j) else backend.matrices.I()
                for j in range(nqubits)
            ),
        )
        for i in range(nqubits)
    )
    return h


def _OneBodyPauli(nqubits, operator, dense: bool = True, backend=None):
    """Helper method for constructing non-interacting
    :math:`X`, :math:`Y`, and :math:`Z` Hamiltonians."""
    backend = _check_backend(backend)
    if dense:
        condition = lambda i, j: i == j % nqubits
        ham = -_build_spin_model(
            nqubits, operator(0, backend=backend).matrix, condition, backend
        )
        return Hamiltonian(nqubits, ham, backend=backend)

    form = sum([-1 * operator(i, backend=backend) for i in range(nqubits)])
    ham = SymbolicHamiltonian(form=form, backend=backend)
    return ham



from typing import Optional

import networkx as nx
import numpy as np

from qibo import gates
from qibo.backends.numpy import NumpyBackend
from qibo.config import raise_error
from qibo.models.circuit import Circuit
from qibo.quantum_info.random_ensembles import random_statevector
from qibo.transpiler._exceptions import (
    ConnectivityError,
    DecompositionError,
    PlacementError,
    TranspilerPipelineError,
)
from qibo.transpiler.optimizer import Preprocessing
from qibo.transpiler.unroller import NativeGates


def assert_transpiling(
    original_circuit: Circuit,
    transpiled_circuit: Circuit,
    connectivity: nx.Graph,
    final_layout: dict,
    native_gates: NativeGates = NativeGates.default(),
    check_circuit_equivalence=True,
):
    """Check that all transpiler passes have been executed correctly.

    Args:
        original_circuit (:class:`qibo.models.circuit.Circuit`): Circuit before transpiling.
        transpiled_circuit (:class:`qibo.models.circuit.Circuit`): Circuit after transpiling.
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
        final_layout (dict): Final {logical: physical} qubit mapping.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`, optional): Native gates supported by the hardware.
            Defaults to :class:`qibo.transpiler.unroller.NativeGates.default()`.
        check_circuit_equivalence (bool, optional): Check if the transpiled circuit is equivalent to the original one.
            Defaults to :math:`True`.
    """
    assert_connectivity(circuit=transpiled_circuit, connectivity=connectivity)
    assert_decomposition(
        circuit=transpiled_circuit,
        native_gates=native_gates,
    )
    if original_circuit.nqubits != transpiled_circuit.nqubits:
        qubit_matcher = Preprocessing(connectivity=connectivity)
        original_circuit = qubit_matcher(circuit=original_circuit)
    assert_placement(circuit=original_circuit, connectivity=connectivity)
    assert_placement(circuit=transpiled_circuit, connectivity=connectivity)
    if check_circuit_equivalence:
        assert_circuit_equivalence(
            original_circuit=original_circuit,
            transpiled_circuit=transpiled_circuit,
            final_layout=final_layout,
        )


def assert_circuit_equivalence(
    original_circuit: Circuit,
    transpiled_circuit: Circuit,
    final_layout: dict,
    test_states: Optional[list] = None,
    ntests: int = 3,
):
    """Checks that the transpiled circuit is equivalent to the original one.

    Args:
        original_circuit (:class:`qibo.models.circuit.Circuit`): Circuit before transpiling.
        transpiled_circuit (:class:`qibo.models.circuit.Circuit`): Circuit after transpiling.
        final_layout (dict): Final {logical: physical} qubit mapping.
        test_states (list, optional): List of states to test the equivalence.
            If ``None``, ``ntests`` random states will be tested. Defauts to ``None``.
        ntests (int, optional): Number of random states to test the equivalence. Defaults to :math: `3`.
    """
    backend = NumpyBackend()
    if transpiled_circuit.nqubits != original_circuit.nqubits:
        raise_error(
            ValueError,
            f"Transpiled circuit ({transpiled_circuit.nqubits}) and original circuit "
            + f"({original_circuit.nqubits}) do not have the same number of qubits.",
        )

    if test_states is None:
        test_states = [
            random_statevector(dims=2**original_circuit.nqubits, backend=backend)
            for _ in range(ntests)
        ]

    ordering = list(final_layout.values())

    for i, state in enumerate(test_states):
        target_state = backend.execute_circuit(
            original_circuit, initial_state=state
        ).state()
        final_state = backend.execute_circuit(
            transpiled_circuit,
            initial_state=state,
        ).state()
        final_state = _transpose_qubits(final_state, ordering)
        fidelity = np.abs(np.dot(np.conj(target_state), final_state))
        try:
            np.testing.assert_allclose(fidelity, 1.0)
        except AssertionError:
            raise_error(TranspilerPipelineError, "Circuit equivalence not satisfied.")


def _transpose_qubits(state: np.ndarray, qubits_ordering: np.ndarray):
    """Reorders qubits of a given state vector.

    Args:
        state (np.ndarray): State vector to reorder.
        qubits_ordering (np.ndarray): Final qubit ordering.
    """
    original_shape = state.shape
    state = np.reshape(state, len(qubits_ordering) * (2,))
    state = np.transpose(state, qubits_ordering)
    return np.reshape(state, original_shape)


def assert_placement(circuit: Circuit, connectivity: nx.Graph):
    """Check if the layout of the circuit is consistent with the circuit and connectivity graph.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to check.
        connectivity (:class:`networkx.Graph`, optional): Hardware connectivity.
    """
    if connectivity is None:
        raise_error(
            ValueError,
            "Connectivity graph is not provided",
        )

    if circuit.nqubits != len(circuit.wire_names) or circuit.nqubits != len(
        connectivity.nodes
    ):
        raise_error(
            PlacementError,
            f"Number of qubits in the circuit ({circuit.nqubits}) "
            + f"does not match either the number of qubits in the layout ({len(circuit.wire_names)}) "
            + f"or the connectivity graph ({len(connectivity.nodes)}).",
        )
    if set(circuit.wire_names) != set(connectivity.nodes):
        raise_error(
            PlacementError,
            "Some physical qubits in the layout may be missing or duplicated.",
        )


def assert_connectivity(connectivity: nx.Graph, circuit: Circuit):
    """Assert if a circuit can be executed on Hardware.

    No gates acting on more than two qubits.
    All two-qubit operations can be performed on hardware.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to check.
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
    """
    layout = circuit.wire_names
    for gate in circuit.queue:
        if len(gate.qubits) > 2 and not isinstance(gate, gates.M):
            raise_error(ConnectivityError, f"{gate.name} acts on more than two qubits.")
        if len(gate.qubits) == 2:
            physical_qubits = (layout[gate.qubits[0]], layout[gate.qubits[1]])
            if physical_qubits not in connectivity.edges:
                raise_error(
                    ConnectivityError,
                    f"The circuit does not respect the connectivity. {gate.name} acts on {physical_qubits} but only the following qubits are directly connected: {connectivity.edges}.",
                )


def assert_decomposition(
    circuit: Circuit,
    native_gates: NativeGates,
):
    """Checks if a circuit has been correctly decomposed into native gates.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to check.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`): Native gates supported by the hardware.
    """
    for gate in circuit.queue:
        if isinstance(gate, gates.M):
            continue
        if len(gate.qubits) <= 2:
            try:
                native_type_gate = NativeGates.from_gate(gate)
                if not native_type_gate & native_gates:
                    raise_error(
                        DecompositionError,
                        f"{gate.name} is not a native gate.",
                    )
            except ValueError:
                raise_error(
                    DecompositionError,
                    f"{gate.name} is not a native gate.",
                )
        else:
            raise_error(
                DecompositionError, f"{gate.name} acts on more than two qubits."
            )
